from __future__ import annotations

import numpy as np
import torch
from torch import Tensor, nn


class Fourier(nn.Module):
    """Fourier Expansion for angle features."""

    def __init__(self, *, order: int = 5, learnable: bool = False) -> None:
        """Initialize the Fourier expansion.

        Args:
            order (int): the maximum order, refer to the N in eq 1 in CHGNet paper
                Default = 5
            learnable (bool): whether to set the frequencies as learnable parameters
                Default = False
        """
        super().__init__()
        self.order = order
        # Initialize frequencies at canonical
        if learnable:
            self.frequencies = torch.nn.Parameter(
                data=torch.arange(1, order + 1, dtype=torch.float),
                requires_grad=True,
            )
        else:
            self.register_buffer(
                "frequencies", torch.arange(1, order + 1, dtype=torch.float)
            )

    def forward(self, x: Tensor) -> Tensor:
        """Apply Fourier expansion to a feature Tensor."""
        result = x.new_zeros(x.shape[0], 1 + 2 * self.order)
        result[:, 0] = 1 / torch.sqrt(torch.tensor([2]))
        tmp = torch.outer(x, self.frequencies)
        result[:, 1 : self.order + 1] = torch.sin(tmp)
        result[:, self.order + 1 :] = torch.cos(tmp)
        return result / np.sqrt(np.pi)


class RadialBessel(torch.nn.Module):
    """1D Bessel Basis
    from: https://github.com/TUM-DAML/gemnet_pytorch/.
    """

    def __init__(
        self,
        *,
        num_radial: int = 9,
        cutoff: float = 5,
        learnable: bool = False,
        smooth_cutoff: int = 5,
    ) -> None:
        """Initialize the SmoothRBF function.

        Args:
            num_radial (int): Controls maximum frequency
                Default = 9
            cutoff (float):  Cutoff distance in Angstrom.
                Default = 5
            learnable (bool): whether to set the frequencies learnable
                Default = False
            smooth_cutoff (int): smooth cutoff strength
                Default = 5
        """
        super().__init__()
        self.num_radial = num_radial
        self.inv_cutoff = 1 / cutoff
        self.norm_const = (2 * self.inv_cutoff) ** 0.5

        # Initialize frequencies at canonical positions
        if learnable:
            self.frequencies = torch.nn.Parameter(
                data=torch.Tensor(
                    np.pi * np.arange(1, self.num_radial + 1, dtype=np.float32)
                ),
                requires_grad=True,
            )
        else:
            self.register_buffer(
                "frequencies",
                np.pi * torch.arange(1, self.num_radial + 1, dtype=torch.float),
            )
        if smooth_cutoff is not None:
            self.smooth_cutoff = CutoffPolynomial(
                cutoff=cutoff, cutoff_coeff=smooth_cutoff
            )
        else:
            self.smooth_cutoff = None

    def forward(
        self, dist: Tensor, *, return_smooth_factor: bool = False
    ) -> Tensor | tuple[Tensor, Tensor]:
        """Apply Bessel expansion to a feature Tensor.

        Args:
            dist (Tensor): tensor of distances [n, 1]
            return_smooth_factor (bool): whether to return the smooth factor
                Default = False

        Returns:
            out (Tensor): tensor of Bessel distances [n, dim]
            where the expanded dimension will be num_radial
            smooth_factor (Tensor): tensor of smooth factors [n, 1]
        """
        dist = dist[:, None]  # shape = (nEdges, 1)
        d_scaled = dist * self.inv_cutoff
        out = self.norm_const * torch.sin(self.frequencies * d_scaled) / dist
        if self.smooth_cutoff is not None:
            smooth_factor = self.smooth_cutoff(dist)
            out = smooth_factor * out
            if return_smooth_factor:
                return out, smooth_factor
        return out


class GaussianExpansion(nn.Module):
    """Expands the distance by Gaussian basis.
    Unit: angstrom.
    """

    def __init__(
        self,
        min: float = 0,  # noqa: A002
        max: float = 5,  # noqa: A002
        step: float = 0.5,
        var: float | None = None,
    ) -> None:
        """Gaussian Expansion
        expand a scalar feature to a soft-one-hot feature vector.

        Args:
            min (float): minimum Gaussian center value
            max (float): maximum Gaussian center value
            step (float): Step size between the Gaussian centers
            var (float): variance in gaussian filter, default to step
        """
        super().__init__()
        if min >= max:
            raise ValueError(f"{min=} must be less than {max=}")
        if max - min <= step:
            raise ValueError(f"{max - min=} must be greater than {step=}")
        self.register_buffer("gaussian_centers", torch.arange(min, max + step, step))
        self.var = var or step
        if self.var <= 0:
            raise ValueError(f"{var=} must be positive")

    def expand(self, features: Tensor) -> Tensor:
        """Apply Gaussian filter to a feature Tensor.

        Args:
            features (Tensor): tensor of features [n]

        Returns:
            expanded features (Tensor): tensor of Gaussian distances [n, dim]
            where the expanded dimension will be (dmax - dmin) / step + 1
        """
        return torch.exp(
            -((features.reshape(-1, 1) - self.gaussian_centers) ** 2) / self.var**2
        )


class CutoffPolynomial(nn.Module):
    """Polynomial soft-cutoff function for atom graph
    ref: https://github.com/TUM-DAML/gemnet_pytorch/blob/-/gemnet/model/layers/envelope.py.
    """

    def __init__(self, cutoff: float = 5, cutoff_coeff: float = 5) -> None:
        """Initialize the polynomial cutoff function.

        Args:
            cutoff (float): cutoff radius (A) in atom graph construction
            Default = 5
            cutoff_coeff (float): the strength of soft-Cutoff
            0 will disable the cutoff, returning 1 at every r
            for positive numbers > 0, the smaller cutoff_coeff is, the faster this
                function decays. Default = 5.
        """
        super().__init__()
        self.cutoff = cutoff
        self.p = cutoff_coeff
        self.a = -(self.p + 1) * (self.p + 2) / 2
        self.b = self.p * (self.p + 2)
        self.c = -self.p * (self.p + 1) / 2

    def forward(self, r: Tensor) -> Tensor:
        """Polynomial cutoff function.

        Args:
            r (Tensor): radius distance tensor

        Returns:
            polynomial cutoff functions: decaying from 1 at r=0 to 0 at r=cutoff
        """
        if self.p != 0:
            r_scaled = r / self.cutoff
            env_val = (
                1
                + self.a * r_scaled**self.p
                + self.b * r_scaled ** (self.p + 1)
                + self.c * r_scaled ** (self.p + 2)
            )
            return torch.where(r_scaled < 1, env_val, torch.zeros_like(r_scaled))
        return r.new_ones(r.shape)



from __future__ import annotations

import itertools
from collections.abc import Sequence

import torch
from torch import Tensor, nn


def aggregate(data: Tensor, owners: Tensor, *, average=True, num_owner=None) -> Tensor:
    """Aggregate rows in data by specifying the owners.

    Args:
        data (Tensor): data tensor to aggregate [n_row, feature_dim]
        owners (Tensor): specify the owner of each row [n_row, 1]
        average (bool): if True, average the rows, if False, sum the rows.
            Default = True
        num_owner (int, optional): the number of owners, this is needed if the
            max idx of owner is not presented in owners tensor
            Default = None

    Returns:
        output (Tensor): [num_owner, feature_dim]
    """
    bin_count = torch.bincount(owners)
    bin_count = bin_count.where(bin_count != 0, bin_count.new_ones(1))

    # If there exist 5 owners, but only the first 4 appear in the owners tensor,
    # we would like the output to have shape [5, fea_dim] with the last row = 0
    # So, we need to optionally add the fifth owner as a row with zero in the output
    if (num_owner is not None) and (bin_count.shape[0] != num_owner):
        difference = num_owner - bin_count.shape[0]
        bin_count = torch.cat([bin_count, bin_count.new_ones(difference)])

    # make sure this operation is done on the same device of data and owners
    output = data.new_zeros([bin_count.shape[0], data.shape[1]])
    output = output.index_add_(0, owners, data)
    if average:
        output = (output.T / bin_count).T
    return output


class MLP(nn.Module):
    """Multi-Layer Perceptron used for non-linear regression."""

    def __init__(
        self,
        input_dim: int,
        *,
        output_dim: int = 1,
        hidden_dim: int | Sequence[int] | None = (64, 64),
        dropout: float = 0,
        activation: str = "silu",
        bias: bool = True,
    ) -> None:
        """Initialize the MLP.

        Args:
            input_dim (int): the input dimension
            output_dim (int): the output dimension
            hidden_dim (list[int] | int]): a list of integers or a single integer
                representing the number of hidden units in each layer of the MLP.
                Default = [64, 64]
            dropout (float): the dropout rate before each linear layer. Default: 0
            activation (str, optional): The name of the activation function to use
                in the gated MLP. Must be one of "relu", "silu", "tanh", or "gelu".
                Default = "silu"
            bias (bool): whether to use bias in each Linear layers.
                Default = True
        """
        super().__init__()
        if hidden_dim is None or hidden_dim == 0:
            layers = [nn.Dropout(dropout), nn.Linear(input_dim, output_dim, bias=bias)]
        elif isinstance(hidden_dim, int):
            layers = [
                nn.Linear(input_dim, hidden_dim, bias=bias),
                find_activation(activation),
                nn.Dropout(dropout),
                nn.Linear(hidden_dim, output_dim, bias=bias),
            ]
        elif isinstance(hidden_dim, Sequence):
            layers = [
                nn.Linear(input_dim, hidden_dim[0], bias=bias),
                find_activation(activation),
            ]
            if len(hidden_dim) != 1:
                for h_in, h_out in itertools.pairwise(hidden_dim):
                    layers.append(nn.Linear(h_in, h_out, bias=bias))
                    layers.append(find_activation(activation))
            layers.append(nn.Dropout(dropout))
            layers.append(nn.Linear(hidden_dim[-1], output_dim, bias=bias))
        else:
            raise TypeError(
                f"{hidden_dim=} must be an integer, a list of integers, or None."
            )
        self.layers = nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        """Performs a forward pass through the MLP.

        Args:
            x (Tensor): a tensor of shape (batch_size, input_dim)

        Returns:
            Tensor: a tensor of shape (batch_size, output_dim)
        """
        return self.layers(x)


class GatedMLP(nn.Module):
    """Gated MLP
    similar model structure is used in CGCNN and M3GNet.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        *,
        hidden_dim: int | list[int] | None = None,
        dropout: float = 0,
        activation: str = "silu",
        norm: str = "batch",
        bias: bool = True,
    ) -> None:
        """Initialize a gated MLP.

        Args:
            input_dim (int): the input dimension
            output_dim (int): the output dimension
            hidden_dim (list[int] | int]): a list of integers or a single integer
                representing the number of hidden units in each layer of the MLP.
                Default = None
            dropout (float): the dropout rate before each linear layer.
                Default: 0
            activation (str, optional): The name of the activation function to use in
                the gated MLP. Must be one of "relu", "silu", "tanh", or "gelu".
                Default = "silu"
            norm (str, optional): The name of the normalization layer to use on the
                updated atom features. Must be one of "batch", "layer", or None.
                Default = "batch"
            bias (bool): whether to use bias in each Linear layers.
                Default = True
        """
        super().__init__()
        self.mlp_core = MLP(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            dropout=dropout,
            activation=activation,
            bias=bias,
        )
        self.mlp_gate = MLP(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_dim=hidden_dim,
            dropout=dropout,
            activation=activation,
            bias=bias,
        )
        self.activation = find_activation(activation)
        self.sigmoid = nn.Sigmoid()
        self.norm = norm
        self.bn1 = find_normalization(name=norm, dim=output_dim)
        self.bn2 = find_normalization(name=norm, dim=output_dim)

    def forward(self, x: Tensor) -> Tensor:
        """Performs a forward pass through the MLP.

        Args:
            x (Tensor): a tensor of shape (batch_size, input_dim)

        Returns:
            Tensor: a tensor of shape (batch_size, output_dim)
        """
        if self.norm is None:
            core = self.activation(self.mlp_core(x))
            gate = self.sigmoid(self.mlp_gate(x))
        else:
            core = self.activation(self.bn1(self.mlp_core(x)))
            gate = self.sigmoid(self.bn2(self.mlp_gate(x)))
        return core * gate


class ScaledSiLU(torch.nn.Module):
    """Scaled Sigmoid Linear Unit."""

    def __init__(self) -> None:
        """Initialize a scaled SiLU."""
        super().__init__()
        self.scale_factor = 1 / 0.6
        self._activation = torch.nn.SiLU()

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass."""
        return self._activation(x) * self.scale_factor


def find_activation(name: str) -> nn.Module:
    """Return an activation function using name."""
    try:
        return {
            "relu": nn.ReLU,
            "silu": nn.SiLU,
            "scaledsilu": ScaledSiLU,
            "gelu": nn.GELU,
            "softplus": nn.Softplus,
            "sigmoid": nn.Sigmoid,
            "tanh": nn.Tanh,
        }[name.lower()]()
    except KeyError as exc:
        raise NotImplementedError from exc


def find_normalization(name: str, dim: int | None = None) -> nn.Module | None:
    """Return an normalization function using name."""
    if name is None:
        return None
    return {
        "batch": nn.BatchNorm1d(dim),
        "layer": nn.LayerNorm(dim),
    }.get(name.lower(), None)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
import torch

from chgnet.utils import cuda_devices_sorted_by_free_mem, solve_charge_by_mag

if TYPE_CHECKING:
    from pymatgen.core import Structure


def test_cuda_devices_sorted_by_free_mem():
    # can't test this any better on CPU
    # but good to check it at least doesn't crash on CPU
    lst = cuda_devices_sorted_by_free_mem()
    if torch.cuda.is_available():
        assert len(lst) > 0
    else:
        assert lst == []


@pytest.mark.parametrize("key", ["final_magmom", "magmom"])
def test_solve_charge_by_mag(li_mn_o2: Structure, key: str) -> None:
    assert li_mn_o2.charge == 0

    li_mn_o2.add_site_property(key, [0.5] * len(li_mn_o2))  # add unphysical magmoms
    # get charge-decorated structure
    struct_with_chg = solve_charge_by_mag(li_mn_o2)
    assert struct_with_chg.charge == -2  # expect unphysical charge



from __future__ import annotations

import os
import pickle
from typing import TYPE_CHECKING, Literal

import numpy as np
import pytest
from ase import Atoms
from ase.md.npt import NPT
from ase.md.nptberendsen import Inhomogeneous_NPTBerendsen
from ase.md.nvtberendsen import NVTBerendsen
from ase.md.verlet import VelocityVerlet
from numpy.testing import assert_allclose
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from pytest import MonkeyPatch, approx  # noqa: PT013

from chgnet import ROOT
from chgnet.graph import CrystalGraphConverter
from chgnet.model import StructOptimizer
from chgnet.model.dynamics import CHGNetCalculator, EquationOfState, MolecularDynamics
from chgnet.model.model import CHGNet

if TYPE_CHECKING:
    from pathlib import Path

relaxer = StructOptimizer()
structure = Structure.from_file(f"{ROOT}/examples/mp-18767-LiMnO2.cif")
structure_2 = Structure.from_file(f"{ROOT}/examples/mp-1175469-Li9Co7O16.cif")
chgnet = CHGNet.load()


def test_version_and_params():
    calculator = relaxer.calculator
    assert isinstance(calculator, CHGNetCalculator)
    model = calculator.model
    assert isinstance(model, CHGNet)
    assert relaxer.version == calculator.version == model.version
    assert relaxer.n_params == calculator.n_params == model.n_params


def test_eos():
    eos = EquationOfState()
    eos.fit(atoms=structure)

    assert eos.get_bulk_modulus() == approx(0.6481, rel=1e-4)
    assert eos.get_bulk_modulus(unit="GPa") == approx(103.845, rel=1e-4)
    assert eos.get_compressibility() == approx(1.5428, rel=1e-4)
    assert eos.get_compressibility(unit="GPa^-1") == approx(0.0096296, rel=1e-4)


@pytest.mark.parametrize("algorithm", ["legacy", "fast"])
def test_md_nvt_berendsen(
    tmp_path: Path, monkeypatch: MonkeyPatch, algorithm: Literal["legacy", "fast"]
):
    monkeypatch.chdir(tmp_path)  # run MD in temporary directory

    chgnet_legacy = CHGNet.load()
    converter_legacy = CrystalGraphConverter(
        atom_graph_cutoff=5, bond_graph_cutoff=3, algorithm=algorithm
    )
    assert converter_legacy.algorithm == algorithm

    chgnet_legacy.graph_converter = converter_legacy

    md = MolecularDynamics(
        atoms=structure,
        model=chgnet_legacy,
        ensemble="nvt",
        thermostat="Berendsen_inhomogeneous",
        temperature=1000,  # in k
        timestep=2,  # in fs
        trajectory="md_out.traj",
        logfile="md_out.log",
        loginterval=10,
    )
    md.run(100)

    assert isinstance(md.atoms, Atoms)
    assert isinstance(md.atoms.calc, CHGNetCalculator)
    assert isinstance(md.dyn, NVTBerendsen)
    assert set(os.listdir()) == {"md_out.log", "md_out.traj"}
    with open("md_out.log") as log_file:
        next(log_file)
        logs = log_file.read()
        logs = np.fromstring(logs, sep=" ")
    ref = np.fromstring(
        "0.0000         -58.8678     -58.8678       0.0000     0.0\n"
        "0.0200         -58.8665     -58.8692       0.0027     2.6\n"
        "0.0400         -58.8650     -58.8846       0.0196    18.9\n"
        "0.0600         -58.7870     -58.8671       0.0801    77.5\n"
        "0.0800         -58.7024     -58.8023       0.0999    96.7\n"
        "0.1000         -58.6080     -58.6803       0.0723    69.9\n"
        "0.1200         -58.5487     -58.5849       0.0362    35.0\n"
        "0.1400         -58.4648     -58.5285       0.0637    61.6\n"
        "0.1600         -58.3202     -58.5693       0.2491   240.9\n"
        "0.1800         -58.2515     -58.7861       0.5346   517.0\n"
        "0.2000         -58.2441     -58.8199       0.5758   556.8\n",
        sep=" ",
    )
    assert_allclose(logs, ref, rtol=2.1e-3, atol=1e-8)


def test_md_nve(tmp_path: Path, monkeypatch: MonkeyPatch):
    monkeypatch.chdir(tmp_path)  # run MD in temporary directory

    md = MolecularDynamics(
        atoms=structure,
        model=chgnet,
        ensemble="nve",
        timestep=1,  # in fs
        trajectory="md_out.traj",
        logfile="md_out.log",
        loginterval=10,
    )
    md.run(10)

    assert isinstance(md.atoms, Atoms)
    assert isinstance(md.atoms.calc, CHGNetCalculator)
    assert isinstance(md.dyn, VelocityVerlet)
    assert set(os.listdir()) == {"md_out.log", "md_out.traj"}
    with open("md_out.log") as log_file:
        logs = log_file.read()
    assert logs == (
        "Time[ps]      Etot[eV]     Epot[eV]     Ekin[eV]    T[K]\n"
        "0.0000         -58.9415     -58.9415       0.0000     0.0\n"
        "0.0100         -58.9415     -58.9417       0.0002     0.2\n"
    )


def test_md_npt_inhomogeneous_berendsen(tmp_path: Path, monkeypatch: MonkeyPatch):
    monkeypatch.chdir(tmp_path)  # run MD in temporary directory

    md = MolecularDynamics(
        atoms=structure,
        model=chgnet,
        ensemble="npt",
        thermostat="Berendsen_inhomogeneous",
        temperature=1000,  # in k
        timestep=2,  # in fs
        trajectory="md_out.traj",
        logfile="md_out.log",
        loginterval=10,
    )
    md.run(100)

    assert isinstance(md.atoms, Atoms)
    assert isinstance(md.atoms.calc, CHGNetCalculator)
    assert isinstance(md.dyn, Inhomogeneous_NPTBerendsen)
    assert md.bulk_modulus == approx(104.16, rel=1e-2)
    assert md.dyn.pressure == approx(6.324e-07, rel=1e-4)
    assert set(os.listdir()) == {"md_out.log", "md_out.traj"}
    with open("md_out.log") as log_file:
        next(log_file)
        logs = log_file.read()
        logs = np.fromstring(logs, sep=" ")
    ref = np.fromstring(
        "0.0000         -58.9415     -58.9415       0.0000     0.0\n"
        "0.0200         -58.9407     -58.9423       0.0016     1.6\n"
        "0.0400         -58.9310     -58.9415       0.0105    10.1\n"
        "0.0600         -58.8819     -58.9315       0.0495    47.9\n"
        "0.0800         -58.7860     -58.8800       0.0940    90.9\n"
        "0.1000         -58.6916     -58.7694       0.0778    75.2\n"
        "0.1200         -58.5945     -58.6458       0.0513    49.6\n"
        "0.1400         -58.4972     -58.5543       0.0571    55.2\n"
        "0.1600         -58.4008     -58.5540       0.1532   148.1\n"
        "0.1800         -58.3292     -58.8330       0.5038   487.2\n"
        "0.2000         -58.2842     -58.8526       0.5684   549.7\n",
        sep=" ",
    )
    assert_allclose(logs, ref, rtol=2.1e-3, atol=1e-8)


def test_md_nvt_nose_hoover(tmp_path: Path, monkeypatch: MonkeyPatch):
    monkeypatch.chdir(tmp_path)  # run MD in temporary directory

    md = MolecularDynamics(
        atoms=structure_2,
        model=chgnet,
        ensemble="nvt",
        thermostat="nose-hoover",
        temperature=1000,  # in k
        timestep=2,  # in fs
        trajectory="md_out.traj",
        logfile="md_out.log",
        loginterval=10,
    )

    new_atoms = AseAtomsAdaptor.get_structure(md.atoms)
    assert StructureMatcher(
        primitive_cell=False, scale=False, attempt_supercell=False, allow_subset=False
    ).fit(structure_2, new_atoms)

    md.run(100)

    assert isinstance(md.atoms, Atoms)
    assert isinstance(md.atoms.calc, CHGNetCalculator)
    assert isinstance(md.dyn, NPT)
    assert_allclose(
        md.dyn.externalstress,
        [-6.324e-07, -6.324e-07, -6.324e-07, 0.0, 0.0, 0.0],
        rtol=1e-4,
        atol=1e-8,
    )
    assert set(os.listdir()) == {"md_out.log", "md_out.traj"}
    with open("md_out.log") as log_file:
        next(log_file)
        logs = log_file.read()
        logs = np.fromstring(logs, sep=" ")
    ref = np.fromstring(
        "0.0200       -199.3047   -199.3890       0.0844    20.4\n"
        "0.0400       -199.3036   -199.3510       0.0475    11.5\n"
        "0.0600       -199.2999   -199.3219       0.0221     5.3\n"
        "0.0800       -199.2974   -199.4012       0.1038    25.1\n"
        "0.1000       -199.2927   -199.3097       0.0170     4.1\n"
        "0.1200       -199.2847   -199.3522       0.0675    16.3\n"
        "0.1400       -199.2802   -199.3789       0.0988    23.9\n"
        "0.1600       -199.2681   -199.2785       0.0104     2.5\n"
        "0.1800       -199.2565   -199.3830       0.1265    30.6\n"
        "0.2000       -199.2463   -199.3190       0.0727    17.6\n",
        sep=" ",
    )
    assert_allclose(logs, ref, rtol=1e-2, atol=1e-7)


def test_md_npt_nose_hoover(tmp_path: Path, monkeypatch: MonkeyPatch):
    # https://github.com/CederGroupHub/chgnet/pull/68
    monkeypatch.chdir(tmp_path)  # run MD in temporary directory

    md = MolecularDynamics(
        atoms=structure_2,
        model=chgnet,
        ensemble="npt",
        thermostat="nose-hoover",
        temperature=1000,  # in k
        timestep=2,  # in fs
        trajectory="md_out.traj",
        logfile="md_out.log",
        loginterval=10,
    )

    new_atoms = AseAtomsAdaptor.get_structure(md.atoms)
    assert StructureMatcher(
        primitive_cell=False, scale=False, attempt_supercell=False, allow_subset=False
    ).fit(structure_2, new_atoms)

    md.run(100)

    assert isinstance(md.atoms, Atoms)
    assert isinstance(md.atoms.calc, CHGNetCalculator)
    assert isinstance(md.dyn, NPT)
    assert md.bulk_modulus == approx(91.14753, rel=1e-2)
    assert_allclose(
        md.dyn.externalstress,
        [-6.324e-07, -6.324e-07, -6.324e-07, 0.0, 0.0, 0.0],
        rtol=1e-4,
        atol=1e-8,
    )
    assert set(os.listdir()) == {"md_out.log", "md_out.traj"}
    with open("md_out.log") as log_file:
        next(log_file)
        logs = log_file.read()
        logs = np.fromstring(logs, sep=" ")
    ref = np.fromstring(
        "0.0200       -199.3048   -199.3891       0.0843    20.4\n"
        "0.0400       -199.3038   -199.3513       0.0475    11.5\n"
        "0.0600       -199.3005   -199.3226       0.0221     5.4\n"
        "0.0800       -199.2988   -199.4024       0.1036    25.0\n"
        "0.1000       -199.2945   -199.3115       0.0170     4.1\n"
        "0.1200       -199.2872   -199.3550       0.0679    16.4\n"
        "0.1400       -199.2841   -199.3822       0.0981    23.7\n"
        "0.1600       -199.2729   -199.2833       0.0105     2.5\n"
        "0.1800       -199.2622   -199.3895       0.1273    30.8\n"
        "0.2000       -199.2539   -199.3247       0.0708    17.1\n",
        sep=" ",
    )

    assert_allclose(logs, ref, rtol=1e-2, atol=1e-7)


def test_md_crystal_feas_log(tmp_path: Path, monkeypatch: MonkeyPatch):
    monkeypatch.chdir(tmp_path)  # run MD in temporary directory

    md = MolecularDynamics(
        atoms=structure,
        ensemble="nvt",
        temperature=1000,  # in k
        timestep=2,  # in fs
        trajectory="md_out.traj",
        logfile="md_out.log",
        crystal_feas_logfile=(logfile := "md_crystal_feas.pkl"),
        loginterval=1,
    )
    md.run(100)

    assert os.path.isfile(logfile)
    with open(logfile, "rb") as file:
        data = pickle.load(file)

    crystal_feas = data["crystal_feas"]

    assert isinstance(crystal_feas, list)
    assert len(crystal_feas) == 101
    assert len(crystal_feas[0]) == 64
    assert crystal_feas[0][0] == approx(-0.002082636, abs=1e-5)
    assert crystal_feas[0][1] == approx(-1.4285042, abs=1e-5)
    assert crystal_feas[10][0] == approx(-0.0020592688, abs=1e-5)
    assert crystal_feas[10][1] == approx(-1.4284436, abs=1e-5)



from __future__ import annotations

import pytest
import torch

from chgnet.model.basis import Fourier, RadialBessel
from chgnet.model.encoders import AngleEncoder, AtomEmbedding, BondEncoder

center = torch.tensor([[1, 0, 0], [0, 0, 0]])
neighbor = torch.tensor([[0, 0, 0], [1, 0, 0]])

bond_i = torch.tensor([[1, 0, 0], [0, 1, 0]])
bond_j = torch.tensor([[0, 1, 0], [1, 0, 0]])


@pytest.mark.parametrize("atom_feature_dim", [16, 32, 64])
@pytest.mark.parametrize("max_num_elements", [94, 89])
def test_atom_embedding(atom_feature_dim: int, max_num_elements: int) -> None:
    atom_embedding = AtomEmbedding(atom_feature_dim, max_num_elements=max_num_elements)
    atomic_numbers = torch.tensor([6, 7, 8])

    result = atom_embedding(atomic_numbers)
    assert result.shape == (3, atom_feature_dim)

    with pytest.raises(IndexError) as exc_info:
        atom_embedding(torch.tensor(100))

    assert "index out of range" in str(exc_info.value)


@pytest.mark.parametrize(("atom_graph_cutoff", "bond_graph_cutoff"), [(5, 3), (6, 4)])
def test_bond_encoder(atom_graph_cutoff: float, bond_graph_cutoff: float) -> None:
    undirected2directed = torch.tensor([0, 1])
    image = torch.zeros((2, 3))
    lattice = torch.eye(3)

    bond_encoder = BondEncoder(
        atom_graph_cutoff=atom_graph_cutoff, bond_graph_cutoff=bond_graph_cutoff
    )
    bond_basis_ag, bond_basis_bg, bond_vectors = bond_encoder(
        center, neighbor, undirected2directed, image, lattice
    )

    assert bond_basis_ag.shape == (2, bond_encoder.rbf_expansion_ag.num_radial)
    assert bond_basis_bg.shape == (2, bond_encoder.rbf_expansion_bg.num_radial)
    assert bond_vectors.shape == (2, 3)


@pytest.mark.parametrize("num_angular", [9, 21])
@pytest.mark.parametrize("learnable", [True, False])
def test_angle_encoder(num_angular: int, learnable: bool) -> None:
    angle_encoder = AngleEncoder(num_angular=num_angular, learnable=learnable)
    result = angle_encoder(bond_i, bond_j)

    assert result.shape == (2, num_angular)
    assert isinstance(angle_encoder.fourier_expansion, Fourier)


@pytest.mark.parametrize("num_angular", [-2, 8])
def test_angle_encoder_num_angular(num_angular: int) -> None:
    with pytest.raises(ValueError, match=f"{num_angular=} must be an odd integer"):
        AngleEncoder(num_angular=num_angular)


@pytest.mark.parametrize("learnable", [True, False])
def test_bond_encoder_learnable(learnable: bool) -> None:
    undirected2directed = torch.tensor([0, 1])
    image = torch.zeros((2, 3))
    lattice = torch.eye(3)

    bond_encoder = BondEncoder(learnable=learnable)
    bond_basis_ag, bond_basis_bg, bond_vectors = bond_encoder(
        center, neighbor, undirected2directed, image, lattice
    )

    assert isinstance(bond_encoder.rbf_expansion_ag, RadialBessel)
    assert isinstance(bond_encoder.rbf_expansion_bg, RadialBessel)
    assert bond_basis_ag.shape == (2, bond_encoder.rbf_expansion_ag.num_radial)
    assert bond_basis_bg.shape == (2, bond_encoder.rbf_expansion_bg.num_radial)
    assert bond_vectors.shape == (2, 3)


def test_bond_encoder_zero_bond_length() -> None:
    center = torch.tensor([[0, 0, 0]])
    neighbor = torch.tensor([[0, 0, 0]])
    undirected2directed = torch.tensor([0])
    image = torch.zeros((1, 3))
    lattice = torch.eye(3)

    bond_encoder = BondEncoder()
    bond_basis_ag, bond_basis_bg, bond_vectors = bond_encoder(
        center, neighbor, undirected2directed, image, lattice
    )

    for tensor in (bond_basis_ag, bond_basis_bg, bond_vectors):
        assert tensor.isnan().all()



from __future__ import annotations

import datetime
import inspect
import os
import random
import shutil
import time
from typing import TYPE_CHECKING, Literal, get_args

import numpy as np
import torch
from torch import Tensor, nn
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    ExponentialLR,
    MultiStepLR,
)

from chgnet.model.model import CHGNet
from chgnet.utils import AverageMeter, determine_device, mae, write_json

try:
    import wandb
except ImportError:
    wandb = None


if TYPE_CHECKING:
    from torch.utils.data import DataLoader
    from typing_extensions import Self

    from chgnet import TrainTask

LogFreq = Literal["epoch", "batch"]
LogEachEpoch, LogEachBatch = get_args(LogFreq)


class Trainer:
    """A trainer to train CHGNet using energy, force, stress and magmom."""

    def __init__(
        self,
        model: CHGNet | None = None,
        *,
        targets: TrainTask = "ef",
        energy_loss_ratio: float = 1,
        force_loss_ratio: float = 1,
        stress_loss_ratio: float = 0.1,
        mag_loss_ratio: float = 0.1,
        optimizer: str = "Adam",
        scheduler: str = "CosLR",
        criterion: str = "MSE",
        epochs: int = 50,
        starting_epoch: int = 0,
        learning_rate: float = 1e-3,
        print_freq: int = 100,
        torch_seed: int | None = None,
        data_seed: int | None = None,
        use_device: str | None = None,
        check_cuda_mem: bool = False,
        wandb_path: str | None = None,
        wandb_init_kwargs: dict | None = None,
        extra_run_config: dict | None = None,
        **kwargs,
    ) -> None:
        """Initialize all hyper-parameters for trainer.

        Args:
            model (nn.Module): a CHGNet model
            targets ("ef" | "efs" | "efsm"): The training targets. Default = "ef"
            energy_loss_ratio (float): energy loss ratio in loss function
                Default = 1
            force_loss_ratio (float): force loss ratio in loss function
                Default = 1
            stress_loss_ratio (float): stress loss ratio in loss function
                Default = 0.1
            mag_loss_ratio (float): magmom loss ratio in loss function
                Default = 0.1
            optimizer (str): optimizer to update model. Can be "Adam", "SGD", "AdamW",
                "RAdam". Default = 'Adam'
            scheduler (str): learning rate scheduler. Can be "CosLR", "ExponentialLR",
                "CosRestartLR". Default = 'CosLR'
            criterion (str): loss function criterion. Can be "MSE", "Huber", "MAE"
                Default = 'MSE'
            epochs (int): number of epochs for training
                Default = 50
            starting_epoch (int): The epoch number to start training at.
            learning_rate (float): initial learning rate
                Default = 1e-3
            print_freq (int): frequency to print training output
                Default = 100
            torch_seed (int): random seed for torch
                Default = None
            data_seed (int): random seed for random
                Default = None
            use_device (str, optional): The device to be used for predictions,
                either "cpu", "cuda", or "mps". If not specified, the default device is
                automatically selected based on the available options.
                Default = None
            check_cuda_mem (bool): Whether to use cuda with most available memory
                Default = False
            wandb_path (str | None): The project and run name separated by a slash:
                "project/run_name". If None, wandb logging is not used.
                Default = None
            wandb_init_kwargs (dict): Additional kwargs to pass to wandb.init.
                Default = None
            extra_run_config (dict): Additional hyper-params to be recorded by wandb
                that are not included in the trainer_args. Default = None

            **kwargs (dict): additional hyper-params for optimizer, scheduler, etc.

        Raises:
            NotImplementedError: If the optimizer or scheduler is not implemented
            ImportError: If wandb_path is specified but wandb is not installed
            ValueError: If wandb_path is specified but not in the format
                'project/run_name'
        """
        # Store trainer args for reproducibility
        self.trainer_args = {
            k: v
            for k, v in locals().items()
            if k not in {"self", "__class__", "model", "kwargs"}
        } | kwargs

        self.model = model
        self.targets = targets

        if torch_seed is not None:
            torch.manual_seed(torch_seed)
        if data_seed:
            random.seed(data_seed)

        # Define optimizer
        if optimizer == "SGD":
            momentum = kwargs.pop("momentum", 0.9)
            weight_decay = kwargs.pop("weight_decay", 0)
            self.optimizer = torch.optim.SGD(
                model.parameters(),
                learning_rate,
                momentum=momentum,
                weight_decay=weight_decay,
            )
        elif optimizer == "Adam":
            weight_decay = kwargs.pop("weight_decay", 0)
            self.optimizer = torch.optim.Adam(
                model.parameters(), learning_rate, weight_decay=weight_decay
            )
        elif optimizer == "AdamW":
            weight_decay = kwargs.pop("weight_decay", 1e-2)
            self.optimizer = torch.optim.AdamW(
                model.parameters(), learning_rate, weight_decay=weight_decay
            )
        elif optimizer == "RAdam":
            weight_decay = kwargs.pop("weight_decay", 0)
            self.optimizer = torch.optim.RAdam(
                model.parameters(), learning_rate, weight_decay=weight_decay
            )

        # Define learning rate scheduler
        default_decay_frac = 1e-2
        if scheduler in {"MultiStepLR", "multistep"}:
            scheduler_params = kwargs.pop(
                "scheduler_params",
                {
                    "milestones": [4 * epochs, 6 * epochs, 8 * epochs, 9 * epochs],
                    "gamma": 0.3,
                },
            )
            self.scheduler = MultiStepLR(self.optimizer, **scheduler_params)
            self.scheduler_type = "multistep"
        elif scheduler in {"ExponentialLR", "Exp", "Exponential"}:
            scheduler_params = kwargs.pop("scheduler_params", {"gamma": 0.98})
            self.scheduler = ExponentialLR(self.optimizer, **scheduler_params)
            self.scheduler_type = "exp"
        elif scheduler in {"CosineAnnealingLR", "CosLR", "Cos", "cos"}:
            scheduler_params = kwargs.pop(
                "scheduler_params", {"decay_fraction": default_decay_frac}
            )
            decay_fraction = scheduler_params.pop("decay_fraction", default_decay_frac)
            self.scheduler = CosineAnnealingLR(
                self.optimizer,
                T_max=10 * epochs,  # Maximum number of iterations.
                eta_min=decay_fraction * learning_rate,
            )
            self.scheduler_type = "cos"
        elif scheduler == "CosRestartLR":
            scheduler_params = kwargs.pop(
                "scheduler_params",
                {"decay_fraction": default_decay_frac, "T_0": 10, "T_mult": 2},
            )
            decay_fraction = scheduler_params.pop("decay_fraction", default_decay_frac)
            self.scheduler = CosineAnnealingWarmRestarts(
                self.optimizer,
                eta_min=decay_fraction * learning_rate,
                **scheduler_params,
            )
            self.scheduler_type = "cosrestart"
        else:
            raise NotImplementedError

        # Define loss criterion
        self.criterion = CombinedLoss(
            target_str=self.targets,
            criterion=criterion,
            is_intensive=self.model.is_intensive,
            energy_loss_ratio=energy_loss_ratio,
            force_loss_ratio=force_loss_ratio,
            stress_loss_ratio=stress_loss_ratio,
            mag_loss_ratio=mag_loss_ratio,
            **kwargs,
        )
        self.epochs = epochs
        self.starting_epoch = starting_epoch

        # Determine the device to use
        self.device = determine_device(
            use_device=use_device, check_cuda_mem=check_cuda_mem
        )

        self.print_freq = print_freq
        self.training_history: dict[
            str, dict[Literal["train", "val", "test"], list[float]]
        ] = {key: {"train": [], "val": [], "test": []} for key in self.targets}
        self.best_model = None

        # Initialize wandb if project/run specified
        if wandb_path:
            if wandb is None:
                raise ImportError(
                    "Weights and Biases not installed. pip install wandb to use "
                    "wandb logging."
                )
            if wandb_path.count("/") == 1:
                project, run_name = wandb_path.split("/")
            else:
                raise ValueError(
                    f"{wandb_path=} should be in the format 'project/run_name' "
                    "(no extra slashes)"
                )
            wandb.init(
                project=project,
                name=run_name,
                config=self.trainer_args | (extra_run_config or {}),
                **(wandb_init_kwargs or {}),
            )

    def train(
        self,
        train_loader: DataLoader,
        val_loader: DataLoader,
        test_loader: DataLoader | None = None,
        *,
        save_dir: str | None = None,
        save_test_result: bool = False,
        train_composition_model: bool = False,
        wandb_log_freq: LogFreq = LogEachBatch,
    ) -> None:
        """Train the model using torch data_loaders.

        Args:
            train_loader (DataLoader): train loader to update CHGNet weights
            val_loader (DataLoader): val loader to test accuracy after each epoch
            test_loader (DataLoader):  test loader to test accuracy at end of training.
                Can be None.
                Default = None
            save_dir (str): the dir name to save the trained weights
                Default = None
            save_test_result (bool): Whether to save the test set prediction in a JSON
                file. Default = False
            train_composition_model (bool): whether to train the composition model
                (AtomRef), this is suggested when the fine-tuning dataset has large
                elemental energy shift from the pretrained CHGNet, which typically comes
                from different DFT pseudo-potentials.
                Default = False
            wandb_log_freq ("epoch" | "batch"): Frequency of logging to wandb.
                'epoch' logs once per epoch, 'batch' logs after every batch.
                Default = "batch"

        Raises:
            ValueError: If model is not initialized
        """
        if self.model is None:
            raise ValueError("Model needs to be initialized")
        global best_checkpoint  # noqa: PLW0603
        if save_dir is None:
            save_dir = f"{datetime.datetime.now(tz=datetime.timezone.utc):%m-%d-%Y}"

        print(f"Begin Training: using {self.device} device")
        print(f"training targets: {self.targets}")
        self.model.to(self.device)

        # Turn composition model training on / off
        for param in self.model.composition_model.parameters():
            param.requires_grad = train_composition_model

        for epoch in range(self.starting_epoch, self.epochs):
            # train
            train_mae = self._train(train_loader, epoch, wandb_log_freq)
            if "e" in train_mae and train_mae["e"] != train_mae["e"]:
                print("Exit due to NaN")
                break

            # val
            val_mae = self._validate(
                val_loader, is_test=False, wandb_log_freq=wandb_log_freq
            )
            for key in self.targets:
                self.training_history[key]["train"].append(train_mae[key])
                self.training_history[key]["val"].append(val_mae[key])

            if "e" in val_mae and val_mae["e"] != val_mae["e"]:
                print("Exit due to NaN")
                break

            if save_dir:
                self.save_checkpoint(epoch, val_mae, save_dir=save_dir)

            # Log epoch metrics to wandb
            if (
                wandb is not None
                and wandb_log_freq == LogEachEpoch
                and self.trainer_args.get("wandb_path")
            ):
                wandb.log(
                    {f"train_{k}_mae": v for k, v in train_mae.items()}
                    | {f"val_{k}_mae": v for k, v in val_mae.items()}
                    | {"epoch": epoch}
                )

        if test_loader is not None:
            # test best model
            print("---------Evaluate Model on Test Set---------------")
            for file in os.listdir(save_dir):
                if file.startswith("bestE_"):
                    test_file = file
                    best_checkpoint = torch.load(os.path.join(save_dir, test_file))

            self.model.load_state_dict(best_checkpoint["model"]["state_dict"])
            test_mae = self._validate(
                test_loader,
                is_test=True,
                test_result_save_path=save_dir if save_test_result else None,
            )

            for key in self.targets:
                self.training_history[key]["test"] = test_mae[key]
            self.save(filename=os.path.join(save_dir, test_file))

            # Log test metrics to wandb
            if wandb is not None and self.trainer_args.get("wandb_path"):
                wandb.log({f"test_{k}_mae": v for k, v in test_mae.items()})

    def _train(
        self,
        train_loader: DataLoader,
        current_epoch: int,
        wandb_log_freq: LogFreq = LogEachBatch,
    ) -> dict:
        """Train all data for one epoch.

        Args:
            train_loader (DataLoader): train loader to update CHGNet weights
            current_epoch (int): used for resume unfinished training
            wandb_log_freq ("epoch" | "batch"): Frequency of logging to wandb

        Returns:
            dictionary of training errors
        """
        batch_time = AverageMeter()
        data_time = AverageMeter()
        losses = AverageMeter()
        mae_errors = {}
        for target in self.targets:
            mae_errors[target] = AverageMeter()

        # switch to train mode
        self.model.train()

        start = time.perf_counter()  # start timer
        for idx, (graphs, targets) in enumerate(train_loader):
            # measure data loading time
            data_time.update(time.perf_counter() - start)

            # get input
            for g in graphs:
                requires_force = "f" in self.targets
                g.atom_frac_coord.requires_grad = requires_force
            graphs = [g.to(self.device) for g in graphs]
            targets = {k: self.move_to(v, self.device) for k, v in targets.items()}

            # compute output
            prediction = self.model(graphs, task=self.targets)
            combined_loss = self.criterion(targets, prediction)

            losses.update(combined_loss["loss"].data.cpu().item(), len(graphs))
            for key in self.targets:
                mae_errors[key].update(
                    combined_loss[f"{key}_MAE"].cpu().item(),
                    combined_loss[f"{key}_MAE_size"],
                )

            # compute gradient and do SGD step
            self.optimizer.zero_grad()
            combined_loss["loss"].backward()
            self.optimizer.step()

            # adjust learning rate every 1/10 of the epoch
            if idx + 1 in np.arange(1, 11) * len(train_loader) // 10:
                self.scheduler.step()

            # free memory
            del graphs, targets
            del prediction, combined_loss

            # measure elapsed time
            batch_time.update(time.perf_counter() - start)
            start = time.perf_counter()

            if idx == 0 or (idx + 1) % self.print_freq == 0:
                message = (
                    f"Epoch: [{current_epoch}][{idx + 1}/{len(train_loader)}] | "
                    f"Time ({batch_time.avg:.3f})({data_time.avg:.3f}) | "
                    f"Loss {losses.val:.4f}({losses.avg:.4f}) | MAE "
                )
                for key in self.targets:
                    message += (
                        f"{key} {mae_errors[key].val:.3f}({mae_errors[key].avg:.3f})  "
                    )
                print(message)

            # Log train metrics to wandb after each batch if specified
            if (
                wandb is not None
                and wandb_log_freq == "batch"
                and self.trainer_args.get("wandb_path")
            ):
                wandb.log(
                    {f"train_{k}_mae": v.avg for k, v in mae_errors.items()}
                    | {"train_loss": losses.avg, "epoch": current_epoch, "batch": idx}
                )

        return {key: round(err.avg, 6) for key, err in mae_errors.items()}

    def _validate(
        self,
        val_loader: DataLoader,
        *,
        is_test: bool = False,
        test_result_save_path: str | None = None,
        wandb_log_freq: LogFreq = LogEachBatch,
    ) -> dict:
        """Validation or test step.

        Args:
            val_loader (DataLoader): val loader to test accuracy after each epoch
            is_test (bool): whether it's test step
            test_result_save_path (str): path to save test_result
            wandb_log_freq ("epoch" | "batch"): Frequency of logging to wandb.
                'epoch' logs once per epoch, 'batch' logs after every batch.

        Returns:
            dictionary of training errors
        """
        batch_time = AverageMeter()
        losses = AverageMeter()
        mae_errors = {}
        for key in self.targets:
            mae_errors[key] = AverageMeter()

        # switch to evaluate mode
        self.model.eval()

        if is_test:
            test_pred = []

        end = time.perf_counter()
        for ii, (graphs, targets) in enumerate(val_loader):
            if "f" in self.targets or "s" in self.targets:
                for graph in graphs:
                    requires_force = "f" in self.targets
                    graph.atom_frac_coord.requires_grad = requires_force
                graphs = [g.to(self.device) for g in graphs]
                targets = {k: self.move_to(v, self.device) for k, v in targets.items()}
            else:
                with torch.no_grad():
                    graphs = [g.to(self.device) for g in graphs]
                    targets = {
                        k: self.move_to(v, self.device) for k, v in targets.items()
                    }

            # compute output
            prediction = self.model(graphs, task=self.targets)
            combined_loss = self.criterion(targets, prediction)

            losses.update(combined_loss["loss"].data.cpu().item(), len(graphs))
            for key in self.targets:
                mae_errors[key].update(
                    combined_loss[f"{key}_MAE"].cpu().item(),
                    combined_loss[f"{key}_MAE_size"],
                )
            if is_test and test_result_save_path:
                for jj, graph_i in enumerate(graphs):
                    tmp = {
                        "mp_id": graph_i.mp_id,
                        "graph_id": graph_i.graph_id,
                        "energy": {
                            "ground_truth": targets["e"][jj].cpu().detach().tolist(),
                            "prediction": prediction["e"][jj].cpu().detach().tolist(),
                        },
                    }
                    if "f" in self.targets:
                        tmp["force"] = {
                            "ground_truth": targets["f"][jj].cpu().detach().tolist(),
                            "prediction": prediction["f"][jj].cpu().detach().tolist(),
                        }
                    if "s" in self.targets:
                        tmp["stress"] = {
                            "ground_truth": targets["s"][jj].cpu().detach().tolist(),
                            "prediction": prediction["s"][jj].cpu().detach().tolist(),
                        }
                    if "m" in self.targets:
                        if targets["m"][jj] is not None:
                            m_ground_truth = targets["m"][jj].cpu().detach().tolist()
                        else:
                            m_ground_truth = None
                        tmp["mag"] = {
                            "ground_truth": m_ground_truth,
                            "prediction": prediction["m"][jj].cpu().detach().tolist(),
                        }
                    test_pred.append(tmp)

            # free memory
            del graphs, targets
            del prediction, combined_loss

            # measure elapsed time
            batch_time.update(time.perf_counter() - end)
            end = time.perf_counter()

            if (ii + 1) % self.print_freq == 0:
                name = "Test" if is_test else "Val"
                message = (
                    f"{name}: [{ii + 1}/{len(val_loader)}] | "
                    f"Time ({batch_time.avg:.3f}) | "
                    f"Loss {losses.val:.4f}({losses.avg:.4f}) | MAE "
                )
                for key in self.targets:
                    message += (
                        f"{key} {mae_errors[key].val:.3f}({mae_errors[key].avg:.3f})  "
                    )
                print(message)

            # Log val metrics to wandb after each batch if specified
            if (
                wandb is not None
                and not is_test
                and wandb_log_freq == "batch"
                and self.trainer_args.get("wandb_path")
            ):
                wandb.log(
                    {f"val_{k}_mae": v.avg for k, v in mae_errors.items()}
                    | {"val_loss": losses.avg, "batch": ii}
                )

        if is_test:
            message = "**  "
            if test_result_save_path:
                write_json(
                    test_pred, os.path.join(test_result_save_path, "test_result.json")
                )
        else:
            message = "*   "
        for key in self.targets:
            message += f"{key}_MAE ({mae_errors[key].avg:.3f}) \t"
        print(message)

        # Log val metrics to wandb at the end of epoch if specified
        if (
            wandb is not None
            and not is_test
            and wandb_log_freq == LogEachEpoch
            and self.trainer_args.get("wandb_path")
        ):
            wandb.log({f"val_{k}_mae": v.avg for k, v in mae_errors.items()})

        return {k: round(mae_error.avg, 6) for k, mae_error in mae_errors.items()}

    def get_best_model(self) -> CHGNet:
        """Get best model recorded in the trainer.

        Returns:
            CHGNet: the model with lowest validation set energy error
        """
        if self.best_model is None:
            raise RuntimeError("the model needs to be trained first")
        MAE = min(self.training_history["e"]["val"])  # noqa: N806
        print(f"Best model has val {MAE =:.4}")
        return self.best_model

    @property
    def _init_keys(self) -> list[str]:
        return [
            key
            for key in list(inspect.signature(Trainer.__init__).parameters)
            if key not in {"self", "model", "kwargs"}
        ]

    def save(self, filename: str = "training_result.pth.tar") -> None:
        """Save the model, graph_converter, etc."""
        state = {
            "model": self.model.as_dict(),
            "optimizer": self.optimizer.state_dict(),
            "scheduler": self.scheduler.state_dict(),
            "training_history": self.training_history,
            "trainer_args": self.trainer_args,
        }
        torch.save(state, filename)

    def save_checkpoint(self, epoch: int, mae_error: dict, save_dir: str) -> None:
        """Function to save CHGNet trained weights after each epoch.

        Args:
            epoch (int): the epoch number
            mae_error (dict): dictionary that stores the MAEs
            save_dir (str): the directory to save trained weights
        """
        os.makedirs(save_dir, exist_ok=True)

        for fname in os.listdir(save_dir):
            if fname.startswith("epoch"):
                os.remove(os.path.join(save_dir, fname))

        err_str = "_".join(
            f"{key}{f'{mae_error[key] * 1000:.0f}' if key in mae_error else 'NA'}"
            for key in "efsm"
        )
        filename = os.path.join(save_dir, f"epoch{epoch}_{err_str}.pth.tar")
        self.save(filename=filename)

        # save the model if it has minimal val energy error or val force error
        if mae_error["e"] == min(self.training_history["e"]["val"]):
            self.best_model = self.model
            for fname in os.listdir(save_dir):
                if fname.startswith("bestE"):
                    os.remove(os.path.join(save_dir, fname))
            shutil.copyfile(
                filename,
                os.path.join(save_dir, f"bestE_epoch{epoch}_{err_str}.pth.tar"),
            )
        if "f" in self.targets and mae_error["f"] == min(
            self.training_history["f"]["val"]
        ):
            for fname in os.listdir(save_dir):
                if fname.startswith("bestF"):
                    os.remove(os.path.join(save_dir, fname))
            shutil.copyfile(
                filename,
                os.path.join(save_dir, f"bestF_epoch{epoch}_{err_str}.pth.tar"),
            )

    @classmethod
    def load(cls, path: str) -> Self:
        """Load trainer state_dict.

        Args:
            path (str): path to the saved model

        Returns:
            Trainer: the loaded trainer
        """
        state = torch.load(path, map_location=torch.device("cpu"))
        model = CHGNet.from_dict(state["model"])
        print(f"Loaded model params = {sum(p.numel() for p in model.parameters()):,}")
        # drop model from trainer_args if present
        state["trainer_args"].pop("model", None)
        trainer = cls(model=model, **state["trainer_args"])
        trainer.model.to(trainer.device)
        trainer.optimizer.load_state_dict(state["optimizer"])
        trainer.scheduler.load_state_dict(state["scheduler"])
        trainer.training_history = state["training_history"]
        trainer.starting_epoch = len(trainer.training_history["e"]["train"])
        return trainer

    @staticmethod
    def move_to(
        obj: Tensor | list[Tensor], device: torch.device
    ) -> Tensor | list[Tensor]:
        """Move object to device.

        Args:
            obj (Tensor | list[Tensor]): object(s) to move to device
            device (torch.device): device to move object to

        Raises:
            TypeError: if obj is not a tensor or list of tensors

        Returns:
            Tensor | list[Tensor]: moved object(s)
        """
        if torch.is_tensor(obj):
            return obj.to(device)
        if isinstance(obj, list):
            out = []
            for tensor in obj:
                if tensor is not None:
                    out.append(tensor.to(device))
                else:
                    out.append(None)
            return out
        raise TypeError("Invalid type for move_to")


class CombinedLoss(nn.Module):
    """A combined loss function of energy, force, stress and magmom."""

    def __init__(
        self,
        *,
        target_str: str = "ef",
        criterion: str = "MSE",
        is_intensive: bool = True,
        energy_loss_ratio: float = 1,
        force_loss_ratio: float = 1,
        stress_loss_ratio: float = 0.1,
        mag_loss_ratio: float = 0.1,
        delta: float = 0.1,
    ) -> None:
        """Initialize the combined loss.

        Args:
            target_str: the training target label. Can be "e", "ef", "efs", "efsm" etc.
                Default = "ef"
            criterion: loss criterion to use
                Default = "MSE"
            is_intensive (bool): whether the energy label is intensive
                Default = True
            energy_loss_ratio (float): energy loss ratio in loss function
                Default = 1
            force_loss_ratio (float): force loss ratio in loss function
                Default = 1
            stress_loss_ratio (float): stress loss ratio in loss function
                Default = 0.1
            mag_loss_ratio (float): magmom loss ratio in loss function
                Default = 0.1
            delta (float): delta for torch.nn.HuberLoss. Default = 0.1
        """
        super().__init__()
        # Define loss criterion
        if criterion in {"MSE", "mse"}:
            self.criterion = nn.MSELoss()
        elif criterion in {"MAE", "mae", "l1"}:
            self.criterion = nn.L1Loss()
        elif criterion == "Huber":
            self.criterion = nn.HuberLoss(delta=delta)
        else:
            raise NotImplementedError
        self.target_str = target_str
        self.is_intensive = is_intensive
        self.energy_loss_ratio = energy_loss_ratio
        if "f" not in self.target_str:
            self.force_loss_ratio = 0
        else:
            self.force_loss_ratio = force_loss_ratio
        if "s" not in self.target_str:
            self.stress_loss_ratio = 0
        else:
            self.stress_loss_ratio = stress_loss_ratio
        if "m" not in self.target_str:
            self.mag_loss_ratio = 0
        else:
            self.mag_loss_ratio = mag_loss_ratio

    def forward(
        self,
        targets: dict[str, Tensor],
        prediction: dict[str, Tensor],
    ) -> dict[str, Tensor]:
        """Compute the combined loss using CHGNet prediction and labels
        this function can automatically mask out magmom loss contribution of
        data points without magmom labels.

        Args:
            targets (dict): DFT labels
            prediction (dict): CHGNet prediction

        Returns:
            dictionary of all the loss, MAE and MAE_size
        """
        out = {"loss": 0.0}
        # Energy
        if "e" in self.target_str:
            if self.is_intensive:
                out["loss"] += self.energy_loss_ratio * self.criterion(
                    targets["e"], prediction["e"]
                )
                out["e_MAE"] = mae(targets["e"], prediction["e"])
                out["e_MAE_size"] = prediction["e"].shape[0]
            else:
                e_per_atom_target = targets["e"] / prediction["atoms_per_graph"]
                e_per_atom_pred = prediction["e"] / prediction["atoms_per_graph"]
                out["loss"] += self.energy_loss_ratio * self.criterion(
                    e_per_atom_target, e_per_atom_pred
                )
                out["e_MAE"] = mae(e_per_atom_target, e_per_atom_pred)
                out["e_MAE_size"] = prediction["e"].shape[0]

        # Force
        if "f" in self.target_str:
            forces_pred = torch.cat(prediction["f"], dim=0)
            forces_target = torch.cat(targets["f"], dim=0)
            out["loss"] += self.force_loss_ratio * self.criterion(
                forces_target, forces_pred
            )
            out["f_MAE"] = mae(forces_target, forces_pred)
            out["f_MAE_size"] = forces_target.shape[0]

        # Stress
        if "s" in self.target_str:
            stress_pred = torch.cat(prediction["s"], dim=0)
            stress_target = torch.cat(targets["s"], dim=0)
            out["loss"] += self.stress_loss_ratio * self.criterion(
                stress_target, stress_pred
            )
            out["s_MAE"] = mae(stress_target, stress_pred)
            out["s_MAE_size"] = stress_target.shape[0]

        # Mag
        if "m" in self.target_str:
            mag_preds, mag_targets = [], []
            m_mae_size = 0
            for mag_pred, mag_target in zip(prediction["m"], targets["m"], strict=True):
                # exclude structures without magmom labels
                if mag_target is not None:
                    mag_preds.append(mag_pred)
                    mag_targets.append(mag_target)
                    m_mae_size += mag_target.shape[0]
            if mag_targets != []:
                mag_preds = torch.cat(mag_preds, dim=0)
                mag_targets = torch.cat(mag_targets, dim=0)
                out["loss"] += self.mag_loss_ratio * self.criterion(
                    mag_targets, mag_preds
                )
                out["m_MAE"] = mae(mag_targets, mag_preds)
                out["m_MAE_size"] = m_mae_size
            else:
                out["m_MAE"] = torch.zeros([1])
                out["m_MAE_size"] = m_mae_size

        return out



from __future__ import annotations

import numpy as np
import pytest

from chgnet.graph.graph import DirectedEdge, Graph, Node, UndirectedEdge


@pytest.fixture
def graph() -> Graph:
    """Create a graph with 3 nodes and 3 directed edges."""
    nodes = [Node(index=idx) for idx in range(3)]
    graph = Graph(nodes)
    graph.add_edge(0, 1, np.zeros(3), 1)
    graph.add_edge(0, 2, np.zeros(3), 2)
    graph.add_edge(1, 2, np.zeros(3), 3)
    return graph


def test_add_edge(graph: Graph) -> None:
    assert len(graph.directed_edges_list) == 3
    assert len(graph.undirected_edges_list) == 3


def test_adjacency_list(graph: Graph) -> None:
    adj_list, directed2undirected = graph.adjacency_list()
    assert len(adj_list) == 3
    assert len(directed2undirected) == 3
    assert adj_list[0] == [0, 1]
    assert adj_list[1] == [0, 2]
    assert adj_list[2] == [1, 2]
    assert directed2undirected[0] == 0
    assert directed2undirected[1] == 1
    assert directed2undirected[2] == 2


def test_undirected2directed(graph: Graph) -> None:
    undirected2directed = graph.undirected2directed()
    assert len(undirected2directed) == 3
    assert undirected2directed[0] == 0
    assert undirected2directed[1] == 1


def test_as_dict(graph: Graph) -> None:
    graph_dict = graph.as_dict()
    assert len(graph_dict["nodes"]) == 3
    assert len(graph_dict["directed_edges"]) == 0
    assert len(graph_dict["directed_edges_list"]) == 3
    assert len(graph_dict["undirected_edges"]) == 3
    assert len(graph_dict["undirected_edges_list"]) == 3


@pytest.fixture
def bigraph() -> Graph:
    """Create a bi-directional graph with 3 nodes and 4 bi-directed edges."""
    nodes = [Node(index=idx) for idx in range(3)]
    bigraph = Graph(nodes)
    bigraph.add_edge(0, 1, np.zeros(3), 1)
    bigraph.add_edge(0, 2, np.zeros(3), 2)
    bigraph.add_edge(1, 0, np.zeros(3), 1)
    bigraph.add_edge(1, 2, np.zeros(3), 5)
    bigraph.add_edge(1, 1, np.array([0, 0, 1]), 4)
    bigraph.add_edge(1, 1, np.array([0, 0, -1]), 4)
    bigraph.add_edge(2, 0, np.zeros(3), 2)
    bigraph.add_edge(2, 1, np.zeros(3), 5)
    return bigraph


def test_add_biedge(bigraph: Graph) -> None:
    assert len(bigraph.directed_edges_list) == 8
    assert len(bigraph.undirected_edges_list) == 4


def test_line_graph(bigraph: Graph) -> None:
    adj_list, directed2undirected = bigraph.adjacency_list()
    line_adj_list, undirected2directed = bigraph.line_graph_adjacency_list(cutoff=7)
    # adj_list
    assert len(adj_list) == 8
    assert len(directed2undirected) == 8
    assert adj_list[0] == [0, 1]
    assert adj_list[1] == [0, 2]
    assert adj_list[2] == [1, 0]
    assert adj_list[6] == [2, 0]
    assert directed2undirected[0] == 0
    assert directed2undirected[1] == 1
    assert directed2undirected[2] == 0

    # line_adj_list
    assert len(line_adj_list) == 16
    assert line_adj_list[0] == [0, 0, 0, 1, 1]
    assert line_adj_list[5] == [2, 1, 6, 2, 7]
    assert line_adj_list[10] == [1, 3, 4, 0, 2]
    assert len(undirected2directed) == 4
    assert undirected2directed[0] == 0
    assert undirected2directed[1] == 1
    assert undirected2directed[2] == 3


def test_directed_edge(capsys) -> None:
    info = {"image": np.zeros(3), "distance": 1}
    edge = DirectedEdge([0, 1], index=0, info=info)
    undirected = edge.make_undirected(index=0, info=info)
    assert edge == edge  # noqa: PLR0124
    captured = capsys.readouterr()
    expected_message = (
        "!!!!!! the two directed edges are equal but this operation is "
        "not supposed to happen\n"
    )
    assert captured.err == expected_message
    assert edge != undirected
    assert edge.nodes == [0, 1]
    assert edge.index == 0
    assert repr(edge) == f"DirectedEdge(nodes=[0, 1], index=0, {info=})"

    # test hashable
    _ = {edge}


def test_undirected_edge() -> None:
    info = {"image": np.zeros(3), "distance": 1}
    edge = UndirectedEdge([0, 1], index=0, info=info)
    assert repr(edge) == f"UndirectedEdge(nodes=[0, 1], index=0, {info=})"

    # test hashable
    _ = {edge}



from __future__ import annotations

import os.path
from typing import TYPE_CHECKING
from zipfile import ZipFile

import pytest
from pymatgen.core import Structure

from chgnet import ROOT
from chgnet.data.dataset import StructureData
from chgnet.utils import parse_vasp_dir

if TYPE_CHECKING:
    from pathlib import Path


def test_parse_vasp_dir_with_magmoms(tmp_path: Path):
    with ZipFile(f"{ROOT}/tests/files/parse-vasp-with-magmoms.zip") as zip_ref:
        zip_ref.extractall(tmp_path)
    vasp_path = os.path.join(tmp_path, "parse-vasp-with-magmoms")
    dataset_dict = parse_vasp_dir(vasp_path, save_path=f"{tmp_path}/tmp.json")

    assert isinstance(dataset_dict, dict)
    assert len(dataset_dict["structure"]) > 0
    assert len(dataset_dict["uncorrected_total_energy"]) > 0
    assert len(dataset_dict["energy_per_atom"]) > 0
    assert len(dataset_dict["force"]) > 0
    assert len(dataset_dict["magmom"]) == len(dataset_dict["force"])
    assert len(dataset_dict["stress"]) > 0

    for structure in dataset_dict["structure"]:
        assert isinstance(structure, Structure)

    for magmom in dataset_dict["magmom"]:
        assert len(magmom) == len(dataset_dict["structure"][0])


def test_parse_vasp_dir_without_magmoms(tmp_path: Path):
    # using test.zip shared for error repro in
    # https://github.com/CederGroupHub/chgnet/issues/147
    with ZipFile(f"{ROOT}/tests/files/parse-vasp-no-magmoms.zip") as zip_ref:
        zip_ref.extractall(tmp_path)
    vasp_path = os.path.join(tmp_path, "parse-vasp-no-magmoms")
    dataset_dict = parse_vasp_dir(vasp_path)

    assert isinstance(dataset_dict, dict)
    assert len(dataset_dict["structure"]) > 0
    assert len(dataset_dict["uncorrected_total_energy"]) > 0
    assert len(dataset_dict["energy_per_atom"]) > 0
    assert len(dataset_dict["force"]) > 0
    assert len(dataset_dict["magmom"]) == 0
    assert len(dataset_dict["stress"]) > 0

    for structure in dataset_dict["structure"]:
        assert isinstance(structure, Structure)

    for magmom in dataset_dict["magmom"]:
        assert len(magmom) == len(dataset_dict["structure"][0])
        assert all(mag == 0.0 for mag in magmom)


def test_parse_vasp_dir_no_data():
    # test non-existing directory
    with pytest.raises(NotADirectoryError, match="is not a directory"):
        parse_vasp_dir(f"{ROOT}/tests/files/non-existent")

    # test existing directory without VASP files
    with pytest.raises(RuntimeError, match="No data parsed from"):
        parse_vasp_dir(f"{ROOT}/tests/files")


def test_dataset_from_vasp_dir(tmp_path: Path):
    with ZipFile(f"{ROOT}/tests/files/parse-vasp-with-magmoms.zip") as zip_ref:
        zip_ref.extractall(tmp_path)
    vasp_path = os.path.join(tmp_path, "parse-vasp-with-magmoms")
    dataset = StructureData.from_vasp(vasp_path, save_path=f"{tmp_path}/tmp.json")
    assert len(dataset.structures) > 0
    assert isinstance(dataset.structures[0], Structure)
    assert len(dataset.structures) == len(dataset.energies)
    assert len(dataset.structures) == len(dataset.forces)
    assert len(dataset.structures) == len(dataset.stresses)
    assert len(dataset.structures) == len(dataset.magmoms)



from __future__ import annotations

import math
import os
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

import torch
from pymatgen.core import Structure
from torch import Tensor, nn

from chgnet.graph import CrystalGraph, CrystalGraphConverter
from chgnet.graph.crystalgraph import TORCH_DTYPE
from chgnet.model.composition_model import AtomRef
from chgnet.model.encoders import AngleEncoder, AtomEmbedding, BondEncoder
from chgnet.model.functions import MLP, GatedMLP, find_normalization
from chgnet.model.layers import (
    AngleUpdate,
    AtomConv,
    BondConv,
    GraphAttentionReadOut,
    GraphPooling,
)
from chgnet.utils import determine_device

if TYPE_CHECKING:
    from typing_extensions import Self

    from chgnet import PredTask

module_dir = os.path.dirname(os.path.abspath(__file__))


class CHGNet(nn.Module):
    """Crystal Hamiltonian Graph neural Network
    A model that takes in a crystal graph and output energy, force, magmom, stress.
    """

    def __init__(
        self,
        *,
        atom_fea_dim: int = 64,
        bond_fea_dim: int = 64,
        angle_fea_dim: int = 64,
        composition_model: str | nn.Module = "MPtrj",
        num_radial: int = 31,
        num_angular: int = 31,
        n_conv: int = 4,
        atom_conv_hidden_dim: Sequence[int] | int = 64,
        update_bond: bool = True,
        bond_conv_hidden_dim: Sequence[int] | int = 64,
        update_angle: bool = True,
        angle_layer_hidden_dim: Sequence[int] | int = 0,
        conv_dropout: float = 0,
        read_out: str = "ave",
        mlp_hidden_dims: Sequence[int] | int = (64, 64, 64),
        mlp_dropout: float = 0,
        mlp_first: bool = True,
        is_intensive: bool = True,
        non_linearity: Literal["silu", "relu", "tanh", "gelu"] = "silu",
        atom_graph_cutoff: float = 6,
        bond_graph_cutoff: float = 3,
        graph_converter_algorithm: Literal["legacy", "fast"] = "fast",
        cutoff_coeff: int = 8,
        learnable_rbf: bool = True,
        gMLP_norm: str | None = "layer",  # noqa: N803
        readout_norm: str | None = "layer",
        version: str | None = None,
        **kwargs,
    ) -> None:
        """Initialize CHGNet.

        Args:
            atom_fea_dim (int): atom feature vector embedding dimension.
                Default = 64
            bond_fea_dim (int): bond feature vector embedding dimension.
                Default = 64
            angle_fea_dim (int): angle feature vector embedding dimension.
                Default = 64
            bond_fea_dim (int): angle feature vector embedding dimension.
                Default = 64
            composition_model (nn.Module, optional): attach a composition model to
                predict energy or initialize a pretrained linear regression (AtomRef).
                The default 'MPtrj' is the atom reference energy linear regression
                trained on all Materials Project relaxation trajectories
                Default = 'MPtrj'
            num_radial (int): number of radial basis used in bond basis expansion.
                Default = 9
            num_angular (int): number of angular basis used in angle basis expansion.
                Default = 9
            n_conv (int): number of interaction blocks.
                Default = 4
                Note: last interaction block contain only an atom_conv layer
            atom_conv_hidden_dim (List or int): hidden dimensions of
                atom convolution layers.
                Default = 64
            update_bond (bool): whether to use bond_conv_layer in bond graph to
                update bond embeddings
                Default = True.
            bond_conv_hidden_dim (List or int): hidden dimensions of
                bond convolution layers.
                Default = 64
            update_angle (bool): whether to use angle_update_layer to
                update angle embeddings.
                Default = True
            angle_layer_hidden_dim (List or int): hidden dimensions of angle layers.
                Default = 0
            conv_dropout (float): dropout rate in all conv_layers.
                Default = 0
            read_out (str): method for pooling layer, 'ave' for standard
                average pooling, 'attn' for multi-head attention.
                Default = "ave"
            mlp_hidden_dims (int or list): readout multilayer perceptron
                hidden dimensions.
                Default = [64, 64]
            mlp_dropout (float): dropout rate in readout MLP.
                Default = 0.
            is_intensive (bool): whether the energy training label is intensive
                i.e. energy per atom.
                Default = True
            non_linearity ('silu' | 'relu' | 'tanh' | 'gelu'): The name of the
                activation function to use in the gated MLP.
                Default = "silu".
            mlp_first (bool): whether to apply mlp first then pooling.
                if set to True, then CHGNet is essentially calculating energy for each
                atom, them sum them up, this is used for the pretrained model
                Default = True
            atom_graph_cutoff (float): cutoff radius (A) in creating atom_graph,
                this need to be consistent with the value in training dataloader
                Default = 5
            bond_graph_cutoff (float): cutoff radius (A) in creating bond_graph,
                this need to be consistent with value in training dataloader
                Default = 3
            graph_converter_algorithm ('legacy' | 'fast'): algorithm to use
                for converting pymatgen.core.Structure to CrystalGraph.
                'legacy': python implementation of graph creation
                'fast': C implementation of graph creation, this is faster,
                    but will need the cygraph.c file correctly compiled from pip install
                default = 'fast'
            cutoff_coeff (float): cutoff strength used in graph smooth cutoff function.
                the smaller this coeff is, the smoother the basis is
                Default = 5
            learnable_rbf (bool): whether to set the frequencies in rbf and Fourier
                basis functions learnable.
                Default = True
            gMLP_norm (str): normalization layer to use in gate-MLP
                Default = 'layer'
            readout_norm (str): normalization layer to use before readout layer
                Default = 'layer'
            version (str): Pretrained checkpoint version.
            **kwargs: Additional keyword arguments
        """
        # Store model args for reconstruction
        self.model_args = {
            key: val
            for key, val in locals().items()
            if key not in {"self", "__class__", "kwargs"}
        }
        self.model_args.update(kwargs)
        if version:
            self.model_args["version"] = version

        super().__init__()
        self.atom_fea_dim = atom_fea_dim
        self.bond_fea_dim = bond_fea_dim
        self.is_intensive = is_intensive
        self.n_conv = n_conv

        # Optionally, define composition model
        if isinstance(composition_model, nn.Module):
            self.composition_model = composition_model
        elif isinstance(composition_model, str):
            self.composition_model = AtomRef(is_intensive=is_intensive)
            self.composition_model.initialize_from(composition_model)
        else:
            self.composition_model = None

        if self.composition_model is not None:
            # fixed composition_model weights
            for param in self.composition_model.parameters():
                param.requires_grad = False

        # Define Crystal Graph Converter
        self.graph_converter = CrystalGraphConverter(
            atom_graph_cutoff=atom_graph_cutoff,
            bond_graph_cutoff=bond_graph_cutoff,
            algorithm=graph_converter_algorithm,
            verbose=kwargs.pop("converter_verbose", False),
        )

        # Define embedding layers
        self.atom_embedding = AtomEmbedding(atom_feature_dim=atom_fea_dim)
        self.bond_basis_expansion = BondEncoder(
            atom_graph_cutoff=atom_graph_cutoff,
            bond_graph_cutoff=bond_graph_cutoff,
            num_radial=num_radial,
            cutoff_coeff=cutoff_coeff,
            learnable=learnable_rbf,
        )
        self.bond_embedding = nn.Linear(
            in_features=num_radial, out_features=bond_fea_dim, bias=False
        )
        self.bond_weights_ag = nn.Linear(
            in_features=num_radial, out_features=atom_fea_dim, bias=False
        )
        self.bond_weights_bg = nn.Linear(
            in_features=num_radial, out_features=bond_fea_dim, bias=False
        )
        self.angle_basis_expansion = AngleEncoder(
            num_angular=num_angular, learnable=learnable_rbf
        )
        self.angle_embedding = nn.Linear(
            in_features=num_angular, out_features=angle_fea_dim, bias=False
        )

        # Define convolutional layers
        conv_norm = kwargs.pop("conv_norm", None)
        mlp_out_bias = kwargs.pop("mlp_out_bias", False)
        atom_graph_layers = [
            AtomConv(
                atom_fea_dim=atom_fea_dim,
                bond_fea_dim=bond_fea_dim,
                hidden_dim=atom_conv_hidden_dim,
                dropout=conv_dropout,
                activation=non_linearity,
                norm=conv_norm,
                gMLP_norm=gMLP_norm,
                use_mlp_out=True,
                mlp_out_bias=mlp_out_bias,
                resnet=True,
            )
            for _ in range(n_conv)
        ]
        self.atom_conv_layers = nn.ModuleList(atom_graph_layers)

        if update_bond:
            bond_graph_layers = [
                BondConv(
                    atom_fea_dim=atom_fea_dim,
                    bond_fea_dim=bond_fea_dim,
                    angle_fea_dim=angle_fea_dim,
                    hidden_dim=bond_conv_hidden_dim,
                    dropout=conv_dropout,
                    activation=non_linearity,
                    norm=conv_norm,
                    gMLP_norm=gMLP_norm,
                    use_mlp_out=True,
                    mlp_out_bias=mlp_out_bias,
                    resnet=True,
                )
                for _ in range(n_conv - 1)
            ]
            self.bond_conv_layers = nn.ModuleList(bond_graph_layers)
        else:
            self.bond_conv_layers = [None for _ in range(n_conv - 1)]

        if update_angle:
            angle_layers = [
                AngleUpdate(
                    atom_fea_dim=atom_fea_dim,
                    bond_fea_dim=bond_fea_dim,
                    angle_fea_dim=angle_fea_dim,
                    hidden_dim=angle_layer_hidden_dim,
                    dropout=conv_dropout,
                    activation=non_linearity,
                    norm=conv_norm,
                    gMLP_norm=gMLP_norm,
                    resnet=True,
                )
                for _ in range(n_conv - 1)
            ]
            self.angle_layers = nn.ModuleList(angle_layers)
        else:
            self.angle_layers = [None for _ in range(n_conv - 1)]

        # Define readout layer
        self.site_wise = nn.Linear(atom_fea_dim, 1)
        self.readout_norm = find_normalization(readout_norm, dim=atom_fea_dim)
        self.mlp_first = mlp_first
        if mlp_first:
            self.read_out_type = "sum"
            input_dim = atom_fea_dim
            self.pooling = GraphPooling(average=False)
        elif read_out in {"attn", "weighted"}:
            self.read_out_type = "attn"
            num_heads = kwargs.pop("num_heads", 3)
            self.pooling = GraphAttentionReadOut(
                atom_fea_dim, num_head=num_heads, average=True
            )
            input_dim = atom_fea_dim * num_heads
        else:
            self.read_out_type = "ave"
            input_dim = atom_fea_dim
            self.pooling = GraphPooling(average=True)
        if kwargs.pop("final_mlp", "MLP") in {"normal", "MLP"}:
            self.mlp = MLP(
                input_dim=input_dim,
                hidden_dim=mlp_hidden_dims,
                output_dim=1,
                dropout=mlp_dropout,
                activation=non_linearity,
            )
        else:
            self.mlp = nn.Sequential(
                GatedMLP(
                    input_dim=input_dim,
                    hidden_dim=mlp_hidden_dims,
                    output_dim=mlp_hidden_dims[-1],
                    dropout=mlp_dropout,
                    norm=gMLP_norm,
                    activation=non_linearity,
                ),
                nn.Linear(in_features=mlp_hidden_dims[-1], out_features=1),
            )

        version_str = f" v{version}" if version else ""
        print(f"CHGNet{version_str} initialized with {self.n_params:,} parameters")

    @property
    def version(self) -> str | None:
        """Return the version of the loaded checkpoint."""
        return self.model_args.get("version")

    @property
    def n_params(self) -> int:
        """Return the number of parameters in the model."""
        return sum(p.numel() for p in self.parameters())

    def forward(
        self,
        graphs: Sequence[CrystalGraph],
        *,
        task: PredTask = "e",
        return_site_energies: bool = False,
        return_atom_feas: bool = False,
        return_crystal_feas: bool = False,
    ) -> dict[str, Tensor]:
        """Get prediction associated with input graphs
        Args:
            graphs (List): a list of CrystalGraphs
            task (str): the prediction task. One of 'e', 'em', 'ef', 'efs', 'efsm'.
                Default = 'e'
            return_site_energies (bool): whether to return per-site energies,
                only available if self.mlp_first == True
                Default = False
            return_atom_feas (bool): whether to return the atom features before last
                conv layer.
                Default = False
            return_crystal_feas (bool): whether to return crystal feature.
                Default = False
        Returns:
            model output (dict).
        """
        # Optionally, make composition model prediction
        comp_energy = (
            0 if self.composition_model is None else self.composition_model(graphs)
        )

        # Make batched graph
        batched_graph = BatchedGraph.from_graphs(
            graphs,
            bond_basis_expansion=self.bond_basis_expansion,
            angle_basis_expansion=self.angle_basis_expansion,
            compute_stress="s" in task,
        )

        # Pass to model
        prediction = self._compute(
            batched_graph,
            compute_force="f" in task,
            compute_stress="s" in task,
            compute_magmom="m" in task,
            return_site_energies=return_site_energies,
            return_atom_feas=return_atom_feas,
            return_crystal_feas=return_crystal_feas,
        )
        prediction["e"] += comp_energy
        if return_site_energies and self.composition_model is not None:
            site_energy_shifts = self.composition_model.get_site_energies(graphs)
            prediction["site_energies"] = [
                i + j
                for i, j in zip(
                    prediction["site_energies"], site_energy_shifts, strict=True
                )
            ]
        return prediction

    def _compute(
        self,
        g: BatchedGraph,
        *,
        compute_force: bool = False,
        compute_stress: bool = False,
        compute_magmom: bool = False,
        return_site_energies: bool = False,
        return_atom_feas: bool = False,
        return_crystal_feas: bool = False,
    ) -> dict:
        """Get Energy, Force, Stress, Magmom associated with input graphs
        force = - d(Energy)/d(atom_positions)
        stress = 1/V * d(Energy)/d(strain).

        Args:
            g (BatchedGraph): batched graph
            compute_force (bool): whether to compute force.
                Default = False
            compute_stress (bool): whether to compute stress.
                Default = False
            compute_magmom (bool): whether to compute magmom.
                Default = False
            return_site_energies (bool): whether to return per-site energies,
                only available if self.mlp_first == True
                Default = False
            return_atom_feas (bool): whether to return atom features.
                Default = False
            return_crystal_feas (bool): whether to return crystal features.
                Default = False

        Returns:
            prediction (dict): containing the fields:
                e (Tensor) : energy of structures [batch_size, 1]
                f (Tensor) : force on atoms [num_batch_atoms, 3]
                s (Tensor) : stress of structure [3 * batch_size, 3]
                m (Tensor) : magnetic moments of sites [num_batch_atoms, 3]
        """
        prediction = {}
        atoms_per_graph = torch.bincount(g.atom_owners)
        prediction["atoms_per_graph"] = atoms_per_graph

        # Embed Atoms, Bonds and Angles
        atom_feas = self.atom_embedding(
            g.atomic_numbers - 1
        )  # let H be the first embedding column
        bond_feas = self.bond_embedding(g.bond_bases_ag)
        bond_weights_ag = self.bond_weights_ag(g.bond_bases_ag)
        bond_weights_bg = self.bond_weights_bg(g.bond_bases_bg)
        if len(g.angle_bases) != 0:
            angle_feas = self.angle_embedding(g.angle_bases)

        # Message Passing
        for idx, (atom_layer, bond_layer, angle_layer) in enumerate(
            zip(
                self.atom_conv_layers[:-1],
                self.bond_conv_layers,
                self.angle_layers,
                strict=False,
            )
        ):
            # Atom Conv
            atom_feas = atom_layer(
                atom_feas=atom_feas,
                bond_feas=bond_feas,
                bond_weights=bond_weights_ag,
                atom_graph=g.batched_atom_graph,
                directed2undirected=g.directed2undirected,
            )

            # Bond Conv
            if len(g.angle_bases) != 0 and bond_layer is not None:
                bond_feas = bond_layer(
                    atom_feas=atom_feas,
                    bond_feas=bond_feas,
                    bond_weights=bond_weights_bg,
                    angle_feas=angle_feas,
                    bond_graph=g.batched_bond_graph,
                )

                # Angle Update
                if angle_layer is not None:
                    angle_feas = angle_layer(
                        atom_feas=atom_feas,
                        bond_feas=bond_feas,
                        angle_feas=angle_feas,
                        bond_graph=g.batched_bond_graph,
                    )
            if idx == self.n_conv - 2:
                if return_atom_feas:
                    prediction["atom_fea"] = torch.split(
                        atom_feas, atoms_per_graph.tolist()
                    )
                # Compute site-wise magnetic moments
                if compute_magmom:
                    magmom = torch.abs(self.site_wise(atom_feas))
                    prediction["m"] = list(
                        torch.split(magmom.view(-1), atoms_per_graph.tolist())
                    )

        # Last conv layer
        atom_feas = self.atom_conv_layers[-1](
            atom_feas=atom_feas,
            bond_feas=bond_feas,
            bond_weights=bond_weights_ag,
            atom_graph=g.batched_atom_graph,
            directed2undirected=g.directed2undirected,
        )
        if self.readout_norm is not None:
            atom_feas = self.readout_norm(atom_feas)

        # Aggregate nodes and ReadOut
        if self.mlp_first:
            energies = self.mlp(atom_feas)
            energy = self.pooling(energies, g.atom_owners).view(-1)
            if return_site_energies:
                prediction["site_energies"] = torch.split(
                    energies.squeeze(1), atoms_per_graph.tolist()
                )
            if return_crystal_feas:
                prediction["crystal_fea"] = self.pooling(atom_feas, g.atom_owners)
        else:  # ave or attn to create crystal_fea first
            crystal_feas = self.pooling(atom_feas, g.atom_owners)
            energy = self.mlp(crystal_feas).view(-1) * atoms_per_graph
            if return_crystal_feas:
                prediction["crystal_fea"] = crystal_feas

        # Compute force
        if compute_force:
            # Need to retain_graph here, because energy is used in loss function,
            # so its gradient need to be calculated later
            # The graphs of force and stress need to be created for same reason.
            force = torch.autograd.grad(
                energy.sum(), g.atom_positions, create_graph=True, retain_graph=True
            )
            prediction["f"] = [-1 * force_dim for force_dim in force]

        # Compute stress
        if compute_stress:
            stress = torch.autograd.grad(
                energy.sum(), g.strains, create_graph=True, retain_graph=True
            )
            # Convert Stress unit from eV/A^3 to GPa
            scale = 1 / g.volumes * 160.21766208
            stress = [i * j for i, j in zip(stress, scale, strict=False)]
            prediction["s"] = stress

        # Normalize energy if model is intensive
        if self.is_intensive:
            energy /= atoms_per_graph
        prediction["e"] = energy

        return prediction

    def predict_structure(
        self,
        structure: Structure | Sequence[Structure],
        *,
        task: PredTask = "efsm",
        return_site_energies: bool = False,
        return_atom_feas: bool = False,
        return_crystal_feas: bool = False,
        batch_size: int = 16,
    ) -> dict[str, Tensor] | list[dict[str, Tensor]]:
        """Predict from pymatgen.core.Structure.

        Args:
            structure (Structure | Sequence[Structure]): structure or a list of
                structures to predict.
            task (str): can be 'e' 'ef', 'em', 'efs', 'efsm'
                Default = "efsm"
            return_site_energies (bool): whether to return per-site energies.
                Default = False
            return_atom_feas (bool): whether to return atom features.
                Default = False
            return_crystal_feas (bool): whether to return crystal features.
                Default = False
            batch_size (int): batch_size for predict structures.
                Default = 16

        Returns:
            prediction (dict): dict or list of dict containing the fields:
                e (Tensor) : energy of structures float in eV/atom
                f (Tensor) : force on atoms [num_atoms, 3] in eV/A
                s (Tensor) : stress of structure [3, 3] in GPa
                m (Tensor) : magnetic moments of sites [num_atoms, 3] in Bohr
                    magneton mu_B
        """
        if self.graph_converter is None:
            raise ValueError("graph_converter cannot be None!")

        structures = [structure] if isinstance(structure, Structure) else structure

        graphs = [self.graph_converter(struct) for struct in structures]
        return self.predict_graph(
            graphs,
            task=task,
            return_site_energies=return_site_energies,
            return_atom_feas=return_atom_feas,
            return_crystal_feas=return_crystal_feas,
            batch_size=batch_size,
        )

    def predict_graph(
        self,
        graph: CrystalGraph | Sequence[CrystalGraph],
        *,
        task: PredTask = "efsm",
        return_site_energies: bool = False,
        return_atom_feas: bool = False,
        return_crystal_feas: bool = False,
        batch_size: int = 16,
    ) -> dict[str, Tensor] | list[dict[str, Tensor]]:
        """Predict from CrustalGraph.

        Args:
            graph (CrystalGraph | Sequence[CrystalGraph]): CrystalGraph(s) to predict.
            task (str): can be 'e' 'ef', 'em', 'efs', 'efsm'
                Default = "efsm"
            return_site_energies (bool): whether to return per-site energies.
                Default = False
            return_atom_feas (bool): whether to return atom features.
                Default = False
            return_crystal_feas (bool): whether to return crystal features.
                Default = False
            batch_size (int): batch_size for predict structures.
                Default = 16

        Returns:
            prediction (dict): dict or list of dict containing the fields:
                e (Tensor) : energy of structures float in eV/atom
                f (Tensor) : force on atoms [num_atoms, 3] in eV/A
                s (Tensor) : stress of structure [3, 3] in GPa
                m (Tensor) : magnetic moments of sites [num_atoms, 3] in Bohr
                    magneton mu_B
        """
        if not isinstance(graph, CrystalGraph | Sequence):
            raise TypeError(
                f"{type(graph)=} must be CrystalGraph or list of CrystalGraphs"
            )

        model_device = next(self.parameters()).device

        graphs = [graph] if isinstance(graph, CrystalGraph) else graph
        self.eval()
        predictions: list[dict[str, Tensor]] = [{} for _ in range(len(graphs))]
        n_steps = math.ceil(len(graphs) / batch_size)
        for step in range(n_steps):
            prediction = self.forward(
                [
                    g.to(model_device)
                    for g in graphs[batch_size * step : batch_size * (step + 1)]
                ],
                task=task,
                return_site_energies=return_site_energies,
                return_atom_feas=return_atom_feas,
                return_crystal_feas=return_crystal_feas,
            )
            for key in {
                "e",
                "f",
                "s",
                "m",
                "site_energies",
                "atom_fea",
                "crystal_fea",
            } & {*prediction}:
                for idx, tensor in enumerate(prediction[key]):
                    predictions[step * batch_size + idx][key] = (
                        tensor.cpu().detach().numpy()
                    )

        return predictions[0] if len(graphs) == 1 else predictions

    def as_dict(self) -> dict:
        """Return the CHGNet weights and args in a dictionary."""
        return {"state_dict": self.state_dict(), "model_args": self.model_args}

    def todict(self) -> dict:
        """Needed for ASE JSON serialization when saving CHGNet potential to
        trajectory file (https://github.com/CederGroupHub/chgnet/issues/48).
        """
        return {"model_name": type(self).__name__, "model_args": self.model_args}

    @classmethod
    def from_dict(cls, dct: dict, **kwargs) -> Self:
        """Build a CHGNet from a saved dictionary."""
        chgnet = cls(**dct["model_args"], **kwargs)
        chgnet.load_state_dict(dct["state_dict"])
        return chgnet

    @classmethod
    def from_file(cls, path: str, **kwargs) -> Self:
        """Build a CHGNet from a saved file."""
        state = torch.load(path, map_location=torch.device("cpu"), weights_only=False)
        return cls.from_dict(state["model"], **kwargs)

    @classmethod
    def load(
        cls,
        *,
        model_name: str = "0.3.0",
        use_device: str | None = None,
        check_cuda_mem: bool = False,
        verbose: bool = True,
    ) -> Self:
        """Load pretrained CHGNet model.

        Args:
            model_name (str, optional):
                Default = "0.3.0".
            use_device (str, optional): The device to be used for predictions,
                either "cpu", "cuda", or "mps". If not specified, the default device is
                automatically selected based on the available options.
                Default = None
            check_cuda_mem (bool): Whether to use cuda with most available memory
                Default = False
            verbose (bool): whether to print model device information
                Default = True
        Raises:
            ValueError: On unknown model_name.
        """
        checkpoint_path = {
            "0.3.0": "../pretrained/0.3.0/chgnet_0.3.0_e29f68s314m37.pth.tar",
            "0.2.0": "../pretrained/0.2.0/chgnet_0.2.0_e30f77s348m32.pth.tar",
        }.get(model_name)

        if checkpoint_path is None:
            raise ValueError(f"Unknown {model_name=}")

        model = cls.from_file(
            os.path.join(module_dir, checkpoint_path),
            # mlp_out_bias=True is set for backward compatible behavior but in rare
            # cases causes unphysical jumps in bonding energy. see
            # https://github.com/CederGroupHub/chgnet/issues/79
            mlp_out_bias=model_name == "0.2.0",
            version=model_name,
        )

        # Determine the device to use
        device = determine_device(use_device=use_device, check_cuda_mem=check_cuda_mem)

        # Move the model to the specified device
        model = model.to(device)
        if verbose:
            print(f"CHGNet will run on {device}")
        return model


@dataclass
class BatchedGraph:
    """Batched crystal graph for parallel computing.

    Attributes:
        atomic_numbers (Tensor): atomic numbers vector
            [num_batch_atoms]
        bond_bases_ag (Tensor): bond bases vector for atom_graph
            [num_batch_bonds_ag, num_radial]
        bond_bases_bg (Tensor): bond bases vector for atom_graph
            [num_batch_bonds_bg, num_radial]
        angle_bases (Tensor): angle bases vector
            [num_batch_angles, num_angular]
        batched_atom_graph (Tensor) : batched atom graph adjacency list
            [num_batch_bonds, 2]
        batched_bond_graph (Tensor) : bond graph adjacency list
            [num_batch_angles, 3]
        atom_owners (Tensor): graph indices for each atom, used aggregate batched
            graph back to single graph
            [num_batch_atoms]
        directed2undirected (Tensor): the utility tensor used to quickly
            map directed edges to undirected edges in graph
            [num_directed]
        atom_positions (list[Tensor]): cartesian coordinates of the atoms
            from structures
            [[num_atoms_1, 3], [num_atoms_2, 3], ...]
        strains (list[Tensor]): a list of strains that's initialized to be zeros
            [[3, 3], [3, 3], ...]
        volumes (Tensor): the volume of each structure in the batch
            [batch_size]
    """

    atomic_numbers: Tensor
    bond_bases_ag: Tensor
    bond_bases_bg: Tensor
    angle_bases: Tensor
    batched_atom_graph: Tensor
    batched_bond_graph: Tensor
    atom_owners: Tensor
    directed2undirected: Tensor
    atom_positions: Sequence[Tensor]
    strains: Sequence[Tensor]
    volumes: Sequence[Tensor] | Tensor

    @classmethod
    def from_graphs(
        cls,
        graphs: Sequence[CrystalGraph],
        bond_basis_expansion: nn.Module,
        angle_basis_expansion: nn.Module,
        *,
        compute_stress: bool = False,
    ) -> Self:
        """Featurize and assemble a list of graphs.

        Args:
            graphs (list[Tensor]): a list of CrystalGraphs
            bond_basis_expansion (nn.Module): bond basis expansion layer in CHGNet
            angle_basis_expansion (nn.Module): angle basis expansion layer in CHGNet
            compute_stress (bool): whether to compute stress. Default = False

        Returns:
            BatchedGraph: assembled graphs ready for batched CHGNet forward pass
        """
        atomic_numbers, atom_positions = [], []
        strains, volumes = [], []
        bond_bases_ag, bond_bases_bg, angle_bases = [], [], []
        batched_atom_graph, batched_bond_graph = [], []
        directed2undirected = []
        atom_owners = []
        atom_offset_idx = n_undirected = 0

        for graph_idx, graph in enumerate(graphs):
            # Atoms
            n_atom = graph.atomic_number.shape[0]
            atomic_numbers.append(graph.atomic_number)

            # Lattice
            if compute_stress:
                strain = graph.lattice.new_zeros([3, 3], requires_grad=True)
                lattice = graph.lattice @ (
                    torch.eye(3, dtype=TORCH_DTYPE).to(strain.device) + strain
                )
            else:
                strain = None
                lattice = graph.lattice
            volumes.append(
                torch.dot(lattice[0], torch.linalg.cross(lattice[1], lattice[2]))
            )
            strains.append(strain)

            # Bonds
            atom_cart_coords = graph.atom_frac_coord @ lattice
            if graph.atom_graph.dim() == 1:
                # This is to avoid structure with all atoms isolated
                graph.atom_graph = graph.atom_graph.reshape(0, 2)
            bond_basis_ag, bond_basis_bg, bond_vectors = bond_basis_expansion(
                center=atom_cart_coords[graph.atom_graph[:, 0]],
                neighbor=atom_cart_coords[graph.atom_graph[:, 1]],
                undirected2directed=graph.undirected2directed,
                image=graph.neighbor_image,
                lattice=lattice,
            )
            atom_positions.append(atom_cart_coords)
            bond_bases_ag.append(bond_basis_ag)
            bond_bases_bg.append(bond_basis_bg)

            # Indexes
            batched_atom_graph.append(graph.atom_graph + atom_offset_idx)
            directed2undirected.append(graph.directed2undirected + n_undirected)

            # Angles
            # Here we use directed edges to calculate angles, and
            # keep only the undirected graph index in the bond_graph,
            # So the number of columns in bond_graph reduce from 5 to 3
            if len(graph.bond_graph) != 0:
                bond_vecs_i = torch.index_select(
                    bond_vectors, 0, graph.bond_graph[:, 2]
                )
                bond_vecs_j = torch.index_select(
                    bond_vectors, 0, graph.bond_graph[:, 4]
                )
                angle_basis = angle_basis_expansion(bond_vecs_i, bond_vecs_j)
                angle_bases.append(angle_basis)

                bond_graph = graph.bond_graph.new_zeros([graph.bond_graph.shape[0], 3])
                bond_graph[:, 0] = graph.bond_graph[:, 0] + atom_offset_idx
                bond_graph[:, 1] = graph.bond_graph[:, 1] + n_undirected
                bond_graph[:, 2] = graph.bond_graph[:, 3] + n_undirected
                batched_bond_graph.append(bond_graph)

            atom_owners.append(torch.ones(n_atom, requires_grad=False) * graph_idx)
            atom_offset_idx += n_atom
            n_undirected += len(bond_basis_ag)

        # Make Torch Tensors
        atomic_numbers = torch.cat(atomic_numbers, dim=0)
        bond_bases_ag = torch.cat(bond_bases_ag, dim=0)
        bond_bases_bg = torch.cat(bond_bases_bg, dim=0)
        angle_bases = (
            torch.cat(angle_bases, dim=0) if len(angle_bases) != 0 else torch.tensor([])
        )
        batched_atom_graph = torch.cat(batched_atom_graph, dim=0)
        if batched_bond_graph != []:
            batched_bond_graph = torch.cat(batched_bond_graph, dim=0)
        else:  # when bond graph is empty or disabled
            batched_bond_graph = torch.tensor([])
        atom_owners = (
            torch.cat(atom_owners, dim=0).type(torch.int32).to(atomic_numbers.device)
        )
        directed2undirected = torch.cat(directed2undirected, dim=0)
        volumes = torch.tensor(volumes, dtype=TORCH_DTYPE, device=atomic_numbers.device)

        return cls(
            atomic_numbers=atomic_numbers,
            bond_bases_ag=bond_bases_ag,
            bond_bases_bg=bond_bases_bg,
            angle_bases=angle_bases,
            batched_atom_graph=batched_atom_graph,
            batched_bond_graph=batched_bond_graph,
            atom_owners=atom_owners,
            directed2undirected=directed2undirected,
            atom_positions=atom_positions,
            strains=strains,
            volumes=volumes,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import pytest
from pymatgen.core import Lattice, Structure

from chgnet.graph import CrystalGraph
from chgnet.graph.converter import CrystalGraphConverter

if TYPE_CHECKING:
    from collections.abc import Generator

lattice = Lattice.cubic(4)
species = ["Na", "Cl"]
coords = [[0, 0, 0], [0.5, 0.5, 0.5]]
NaCl = Structure(lattice, species, coords)


@pytest.fixture
def _set_make_graph() -> Generator[None, None, None]:
    # fixture to force make_graph to be None and then restore it after test
    from chgnet.graph import converter

    make_graph = converter.make_graph  # save original value
    converter.make_graph = None  # force make_graph to be None
    yield  # allows us to have cleanup after the test
    converter.make_graph = make_graph  # restore original value


@pytest.mark.parametrize(
    ("atom_graph_cutoff", "bond_graph_cutoff"), [(5, 3), (5, None), (4, 2)]
)
def test_crystal_graph_converter_cutoff(
    atom_graph_cutoff: float | None, bond_graph_cutoff: float | None
):
    converter = CrystalGraphConverter(
        atom_graph_cutoff=atom_graph_cutoff, bond_graph_cutoff=bond_graph_cutoff
    )
    assert converter.atom_graph_cutoff == atom_graph_cutoff
    assert converter.bond_graph_cutoff == bond_graph_cutoff or atom_graph_cutoff


@pytest.mark.parametrize("algorithm", ["legacy", "fast"])
def test_crystal_graph_converter_algorithm(algorithm: Literal["legacy", "fast"]):
    converter = CrystalGraphConverter(
        atom_graph_cutoff=5, bond_graph_cutoff=3, algorithm=algorithm
    )
    assert converter.algorithm == algorithm


@pytest.mark.usefixtures("_set_make_graph")
def test_crystal_graph_converter_warns():
    with pytest.warns(UserWarning, match="Unknown algorithm='foobar', using `legacy`"):
        CrystalGraphConverter(
            atom_graph_cutoff=5, bond_graph_cutoff=3, algorithm="foobar"
        )
    with pytest.warns(
        UserWarning, match="`fast` algorithm is not available, using `legacy`"
    ):
        CrystalGraphConverter(
            atom_graph_cutoff=5, bond_graph_cutoff=3, algorithm="fast"
        )


@pytest.mark.parametrize("on_isolated_atoms", ["ignore", "warn", "error"])
def test_crystal_graph_converter_forward(
    on_isolated_atoms, capsys: pytest.CaptureFixture[str]
):
    atom_graph_cutoff = 5
    converter = CrystalGraphConverter(
        atom_graph_cutoff=atom_graph_cutoff,
        bond_graph_cutoff=3,
        on_isolated_atoms=on_isolated_atoms,
    )
    strained = NaCl.copy()
    strained.apply_strain(5)
    graph_id = "strained"
    err_msg = (
        f"Structure {graph_id=} has {len(NaCl)} isolated atom(s) with "
        f"{atom_graph_cutoff=}. "
        f"CHGNet calculation will likely go wrong"
    )

    if on_isolated_atoms == "error":
        with pytest.raises(
            ValueError, match=f"Structure {graph_id=} has {len(NaCl)} isolated atom"
        ) as exc_info:
            converter.forward(strained, graph_id=graph_id)
        assert err_msg in str(exc_info.value)
    else:
        crystal_graph = converter.forward(strained, graph_id=graph_id)
        assert isinstance(crystal_graph, CrystalGraph)
        stdout, stderr = capsys.readouterr()
        assert stdout == ""
        if on_isolated_atoms == "warn":
            assert err_msg in stderr
        else:
            assert stderr == ""


def test_crystal_graph_converter_as_dict_round_trip():
    expected = {"atom_graph_cutoff": 5, "bond_graph_cutoff": 3}
    converter = CrystalGraphConverter(**expected)
    converter2 = CrystalGraphConverter.from_dict(converter.as_dict())
    assert converter.atom_graph_cutoff == converter2.atom_graph_cutoff
    assert converter.bond_graph_cutoff == converter2.bond_graph_cutoff
    assert converter.algorithm == converter2.algorithm



from __future__ import annotations

import os
import re
from typing import Literal

import pytest
import torch
from ase.filters import ExpCellFilter, Filter, FrechetCellFilter
from pymatgen.core import Structure

from chgnet.graph import CrystalGraphConverter
from chgnet.model import CHGNet, StructOptimizer


@pytest.mark.parametrize(
    ("algorithm", "ase_filter", "assign_magmoms"),
    [("legacy", FrechetCellFilter, True), ("fast", ExpCellFilter, False)],
)
def test_relaxation(
    algorithm: Literal["legacy", "fast"],
    ase_filter: Filter,
    assign_magmoms: bool,
    li_mn_o2: Structure,
) -> None:
    chgnet = CHGNet.load()
    converter = CrystalGraphConverter(
        atom_graph_cutoff=6, bond_graph_cutoff=3, algorithm=algorithm
    )
    assert converter.algorithm == algorithm

    chgnet.graph_converter = converter
    relaxer = StructOptimizer(model=chgnet)
    result = relaxer.relax(
        li_mn_o2, verbose=True, ase_filter=ase_filter, assign_magmoms=assign_magmoms
    )
    assert list(result) == ["final_structure", "trajectory"]
    final_struct, traj = result["final_structure"], result["trajectory"]
    assert isinstance(final_struct, Structure)
    if assign_magmoms:
        assert isinstance(final_struct.site_properties["magmom"], list)
        assert len(final_struct.site_properties["magmom"]) == len(final_struct)
        assert all(
            isinstance(mm, float) for mm in final_struct.site_properties["magmom"]
        )
    else:
        assert "magmom" not in final_struct.site_properties

    # make sure trajectory has expected attributes
    assert {*traj.__dict__} == {
        *"atoms energies forces stresses magmoms atom_positions cells".split()
    }
    assert len(traj) == 2 if algorithm == "legacy" else 4

    # make sure final structure is more relaxed than initial one
    assert traj.energies[-1] == pytest.approx(-58.94209, rel=1e-4)


no_cuda = pytest.mark.skipif(not torch.cuda.is_available(), reason="No CUDA device")
# skip in macos-14 M1 CI due to OOM error (TODO investigate if
# PYTORCH_MPS_HIGH_WATERMARK_RATIO can fix)
no_mps = pytest.mark.skipif(
    not torch.backends.mps.is_available() or "CI" in os.environ, reason="No MPS device"
)


@pytest.mark.parametrize(
    "use_device",
    ["cpu", pytest.param("cuda", marks=no_cuda), pytest.param("mps", marks=no_mps)],
)
def test_structure_optimizer_passes_kwargs_to_model(use_device: str) -> None:
    relaxer = StructOptimizer(use_device=use_device)
    assert re.match(rf"{use_device}(:\d+)?", relaxer.calculator.device)



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import numpy as np
import pytest
import torch
import wandb
from pymatgen.core import Lattice, Structure

from chgnet.data.dataset import StructureData, get_train_val_test_loader
from chgnet.model import CHGNet
from chgnet.trainer import Trainer

if TYPE_CHECKING:
    from pathlib import Path

lattice = Lattice.cubic(4)
species = ["Na", "Cl"]
coords = [[0, 0, 0], [0.5, 0.5, 0.5]]
NaCl = Structure(lattice, species, coords)
structures, energies, forces, stresses, magmoms = [], [], [], [], []
for _ in range(100):
    struct = NaCl.copy()
    struct.perturb(0.1)
    structures.append(struct)
    energies.append(np.random.random(1))
    forces.append(np.random.random([2, 3]))
    stresses.append(np.random.random([3, 3]))
    magmoms.append(np.random.random(2))

data = StructureData(
    structures=structures,
    energies=energies,
    forces=forces,
    stresses=stresses,
    magmoms=magmoms,
)
train_loader, val_loader, _test_loader = get_train_val_test_loader(
    data, batch_size=16, train_ratio=0.9, val_ratio=0.05
)
chgnet = CHGNet.load()


def test_trainer(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    extra_run_config = dict(some_other_hyperparam=42)
    trainer = Trainer(
        model=chgnet,
        targets="efsm",
        optimizer="Adam",
        criterion="MSE",
        learning_rate=1e-2,
        epochs=5,
        wandb_path="test/run",
        wandb_init_kwargs=dict(anonymous="must"),
        extra_run_config=extra_run_config,
    )
    trainer.train(
        train_loader,
        val_loader,
        save_dir=tmp_path,
        save_test_result=tmp_path / "test-preds.json",
    )
    assert dict(wandb.config).items() >= extra_run_config.items()
    for param in chgnet.composition_model.parameters():
        assert param.requires_grad is False
    assert tmp_path.is_dir(), "Training dir was not created"

    output_files = [file.name for file in tmp_path.iterdir()]
    for prefix in ("epoch", "bestE_", "bestF_"):
        n_matches = sum(file.startswith(prefix) for file in output_files)
        assert (
            n_matches == 1
        ), f"Expected 1 {prefix} file, found {n_matches} in {output_files}"

    # expect ImportError when passing wandb_path without wandb installed
    err_msg = "Weights and Biases not installed. pip install wandb to use wandb logging"
    with monkeypatch.context() as ctx, pytest.raises(ImportError, match=err_msg):  # noqa: PT012
        ctx.setattr("chgnet.trainer.trainer.wandb", None)
        _ = Trainer(model=chgnet, wandb_path="some-org/some-project")


def test_trainer_composition_model(tmp_path: Path) -> None:
    for param in chgnet.composition_model.parameters():
        assert param.requires_grad is False
    trainer = Trainer(
        model=chgnet,
        targets="efsm",
        optimizer="Adam",
        criterion="MSE",
        learning_rate=1e-2,
        epochs=5,
    )
    initial_weights = chgnet.composition_model.state_dict()["fc.weight"].clone()
    trainer.train(
        train_loader, val_loader, save_dir=tmp_path, train_composition_model=True
    )
    for param in chgnet.composition_model.parameters():
        assert param.requires_grad is True

    output_files = list(tmp_path.iterdir())
    weights_path = next(file for file in output_files if file.name.startswith("epoch"))
    new_chgnet = CHGNet.from_file(weights_path)
    for param in new_chgnet.composition_model.parameters():
        assert param.requires_grad is False
    comparison = new_chgnet.composition_model.state_dict()["fc.weight"].to(
        "cpu"
    ) == initial_weights.to("cpu")
    expect = torch.ones_like(comparison)
    # Only Na and Cl should have updated
    expect[0][10] = 0
    expect[0][16] = 0
    assert torch.all(comparison == expect)


@pytest.fixture
def mock_wandb():
    with patch("chgnet.trainer.trainer.wandb") as mock:
        yield mock


def test_wandb_init(mock_wandb):
    chgnet = CHGNet.load()
    _trainer = Trainer(
        model=chgnet,
        wandb_path="test-project/test-run",
        wandb_init_kwargs={"tags": ["test"]},
    )
    expected_config = {
        "targets": "ef",
        "energy_loss_ratio": 1,
        "force_loss_ratio": 1,
        "stress_loss_ratio": 0.1,
        "mag_loss_ratio": 0.1,
        "optimizer": "Adam",
        "scheduler": "CosLR",
        "criterion": "MSE",
        "epochs": 50,
        "starting_epoch": 0,
        "learning_rate": 0.001,
        "print_freq": 100,
        "torch_seed": None,
        "data_seed": None,
        "use_device": None,
        "check_cuda_mem": False,
        "wandb_path": "test-project/test-run",
        "wandb_init_kwargs": {"tags": ["test"]},
        "extra_run_config": None,
    }
    mock_wandb.init.assert_called_once_with(
        project="test-project", name="test-run", config=expected_config, tags=["test"]
    )


def test_wandb_log_frequency(tmp_path, mock_wandb):
    trainer = Trainer(model=chgnet, wandb_path="test-project/test-run", epochs=1)

    # Test epoch logging
    trainer.train(train_loader, val_loader, wandb_log_freq="epoch", save_dir=tmp_path)
    assert (
        mock_wandb.log.call_count == 2 * trainer.epochs
    ), "Expected one train and one val log per epoch"

    mock_wandb.log.reset_mock()

    # Test batch logging
    trainer.train(train_loader, val_loader, wandb_log_freq="batch", save_dir=tmp_path)
    expected_batch_calls = trainer.epochs * len(train_loader)
    assert (
        mock_wandb.log.call_count > expected_batch_calls
    ), "Expected more calls for batch logging"

    # Test log content (for both epoch and batch logging)
    for call_args in mock_wandb.log.call_args_list:
        logged_data = call_args[0][0]
        assert isinstance(logged_data, dict), "Logged data should be a dictionary"
        assert any(
            key.endswith("_mae") for key in logged_data
        ), "Logged data should contain MAE metrics"

    mock_wandb.log.reset_mock()

    # Test no logging when wandb_path is not provided
    trainer_no_wandb = Trainer(model=chgnet, epochs=1)
    trainer_no_wandb.train(train_loader, val_loader, save_dir=tmp_path)
    mock_wandb.log.assert_not_called()



# cython: language_level=3
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: profile=False
# distutils: language = c


import numpy as np
cimport numpy as np

import chgnet.graph.graph

from libc.stdlib cimport free


cdef extern from 'fast_converter_libraries/create_graph.c':
    ctypedef struct Node:
        np.int64_t index
        LongToDirectedEdgeList* neighbors
        np.int64_t num_neighbors

    ctypedef struct NodeIndexPair:
        np.int64_t center
        np.int64_t neighbor

    ctypedef struct UndirectedEdge:
        NodeIndexPair nodes
        np.int64_t index
        np.int64_t* directed_edge_indices
        np.int64_t num_directed_edges
        np.float64_t distance

    ctypedef struct DirectedEdge:
        NodeIndexPair nodes
        np.int64_t index
        const np.int64_t* image
        np.int64_t undirected_edge_index
        np.float64_t distance

    ctypedef struct LongToDirectedEdgeList:
        np.int64_t key
        DirectedEdge** directed_edges_list
        int num_directed_edges_in_group

    ctypedef struct ReturnElems2:
        np.int64_t num_nodes
        np.int64_t num_directed_edges
        np.int64_t num_undirected_edges
        Node* nodes
        UndirectedEdge** undirected_edges_list
        DirectedEdge** directed_edges_list

    ReturnElems2* create_graph(
        np.int64_t* center_index,
        np.int64_t n_e,
        np.int64_t* neighbor_index,
        np.int64_t* image,
        np.float64_t* distance,
        np.int64_t num_atoms)

    void free_LongToDirectedEdgeList_in_nodes(Node* nodes, np.int64_t num_nodes)


    LongToDirectedEdgeList** get_neighbors(Node* node)

def make_graph(
        const np.int64_t[::1] center_index,
        const np.int64_t n_e,
        const np.int64_t[::1] neighbor_index,
        const np.int64_t[:, ::1] image,
        const np.float64_t[::1] distance,
        const np.int64_t num_atoms
    ):
    cdef ReturnElems2* returned
    returned = <ReturnElems2*> create_graph(<np.int64_t*> &center_index[0], n_e, <np.int64_t*> &neighbor_index[0], <np.int64_t*> &image[0][0], <np.float64_t*> &distance[0], num_atoms)

    chg_DirectedEdge = chgnet.graph.graph.DirectedEdge
    chg_Node = chgnet.graph.graph.Node
    chg_UndirectedEdge = chgnet.graph.graph.UndirectedEdge


    image_np = np.asarray(image)

    cdef LongToDirectedEdgeList** node_neighbors
    cdef Node this_node
    cdef LongToDirectedEdgeList this_entry
    py_nodes = []
    cdef DirectedEdge* this_DE


    # Handling nodes + directed edges
    for idx in range(returned[0].num_nodes):
        this_node = returned[0].nodes[idx]
        this_py_node = chg_Node(index=idx)

        node_neighbors = get_neighbors(&this_node)

        # Iterate through all neighbors and populate our py_node.neighbors dict
        for j in range(this_node.num_neighbors):
            this_entry = node_neighbors[j][0]
            directed_edges = []

            for k in range(this_entry.num_directed_edges_in_group):
                this_DE = this_entry.directed_edges_list[k]
                directed_edges.append(this_DE[0].index)

            this_py_node.neighbors[this_entry.key] = directed_edges

        py_nodes.append(this_py_node)

        free(node_neighbors)

    # Handling directed edges
    py_directed_edges_list = []

    for idx in range(returned[0].num_directed_edges):
        this_DE = returned[0].directed_edges_list[idx]
        py_DE = chg_DirectedEdge(nodes = [this_DE[0].nodes.center, this_DE[0].nodes.neighbor], index=this_DE[0].index, info = {"distance": this_DE[0].distance, "image": image_np[this_DE[0].index], "undirected_edge_index": this_DE[0].undirected_edge_index})

        py_directed_edges_list.append(py_DE)


    # Handling undirected edges
    py_undirected_edges_list = []
    cdef UndirectedEdge* UDE

    for idx in range(returned[0].num_undirected_edges):
        UDE = returned[0].undirected_edges_list[idx]
        py_undirected_edge = chg_UndirectedEdge([UDE[0].nodes.center, UDE[0].nodes.neighbor], index = UDE[0].index, info =  {"distance": UDE[0].distance, "directed_edge_index": []})

        for j in range(UDE[0].num_directed_edges):
            py_undirected_edge.info["directed_edge_index"].append(UDE[0].directed_edge_indices[j])

        py_undirected_edges_list.append(py_undirected_edge)


    # Create Undirected_Edges hashmap
    py_undirected_edges = {}
    for undirected_edge in py_undirected_edges_list:
        this_set = frozenset(undirected_edge.nodes)
        if this_set not in py_undirected_edges:
            py_undirected_edges[this_set] = [undirected_edge]
        else:
            py_undirected_edges[this_set].append(undirected_edge)

    # # Update the nodes list to have pointers to DirectedEdges instead of indices
    for node_index in range(returned[0].num_nodes):
        this_neighbors = py_nodes[node_index].neighbors
        for this_neighbor_index in this_neighbors:
            replacement = [py_directed_edges_list[edge_index] for edge_index in this_neighbors[this_neighbor_index]]
            this_neighbors[this_neighbor_index] = replacement


    # Free everything unneeded
    for idx in range(returned[0].num_directed_edges):
        free(returned[0].directed_edges_list[idx])

    for idx in range(returned[0].num_undirected_edges):
        free(returned[0].undirected_edges_list[idx].directed_edge_indices)
        free(returned[0].undirected_edges_list[idx])


    # Free node LongToDirectedEdgeList
    free_LongToDirectedEdgeList_in_nodes(returned[0].nodes, returned[0].num_nodes)

    free(returned[0].directed_edges_list)
    free(returned[0].undirected_edges_list)
    free(returned[0].nodes)

    free(returned)

    return py_nodes, py_directed_edges_list, py_undirected_edges_list, py_undirected_edges



from __future__ import annotations

import collections
from typing import TYPE_CHECKING

import numpy as np
import torch
from pymatgen.core import Structure
from torch import Tensor, nn

from chgnet.model.functions import GatedMLP, find_activation

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    from chgnet.graph.crystalgraph import CrystalGraph


class CompositionModel(nn.Module):
    """A simple FC model that takes in a chemical composition (no structure info)
    and outputs energy.
    """

    def __init__(
        self,
        *,
        atom_fea_dim: int = 64,
        activation: str = "silu",
        is_intensive: bool = True,
        max_num_elements: int = 94,
    ) -> None:
        """Initialize a CompositionModel."""
        super().__init__()
        self.is_intensive = is_intensive
        self.max_num_elements = max_num_elements
        self.fc1 = nn.Linear(max_num_elements, atom_fea_dim)
        self.activation = find_activation(activation)
        self.gated_mlp = GatedMLP(
            input_dim=atom_fea_dim,
            output_dim=atom_fea_dim,
            hidden_dim=atom_fea_dim,
            activation=activation,
        )
        self.fc2 = nn.Linear(atom_fea_dim, 1)

    def _get_energy(self, composition_feas: Tensor) -> Tensor:
        """Predict the energy given composition encoding.

        Args:
            composition_feas: batched atom feature matrix of shape
                [batch_size, total_num_elements].

        Returns:
            prediction associated with each composition [batchsize].
        """
        composition_feas = self.activation(self.fc1(composition_feas))
        composition_feas += self.gated_mlp(composition_feas)
        return self.fc2(composition_feas).view(-1)

    def forward(self, graphs: list[CrystalGraph]) -> Tensor:
        """Get the energy of a list of CrystalGraphs as Tensor."""
        composition_feas = self._assemble_graphs(graphs)
        return self._get_energy(composition_feas)

    def _assemble_graphs(self, graphs: list[CrystalGraph]) -> Tensor:
        """Assemble a list of graphs into one-hot composition encodings.

        Args:
            graphs (list[CrystalGraph]): a list of CrystalGraphs

        Returns:
            assembled batch_graph that contains all information for model.
        """
        composition_feas = []
        for graph in graphs:
            composition_fea = torch.bincount(
                graph.atomic_number - 1, minlength=self.max_num_elements
            )
            if self.is_intensive:
                n_atom = graph.atomic_number.shape[0]
                composition_fea = composition_fea / n_atom
            composition_feas.append(composition_fea)
        return torch.stack(composition_feas, dim=0)


class AtomRef(nn.Module):
    """A linear regression for elemental energy.
    From: https://github.com/materialsvirtuallab/m3gnet/.
    """

    def __init__(
        self, *, is_intensive: bool = True, max_num_elements: int = 94
    ) -> None:
        """Initialize an AtomRef model."""
        super().__init__()
        self.is_intensive = is_intensive
        self.max_num_elements = max_num_elements
        self.fc = nn.Linear(max_num_elements, 1, bias=False)
        self.fitted = False

    def forward(self, graphs: list[CrystalGraph]) -> Tensor:
        """Get the energy of a list of CrystalGraphs.

        Args:
            graphs (List(CrystalGraph)): a list of Crystal Graph to compute

        Returns:
            energy (tensor)
        """
        if not self.fitted:
            raise ValueError("composition model needs to be fitted first!")
        composition_feas = self._assemble_graphs(graphs)
        return self._get_energy(composition_feas)

    def _get_energy(self, composition_feas: Tensor) -> Tensor:
        """Predict the energy given composition encoding.

        Args:
            composition_feas: batched atom feature matrix of shape
                [batch_size, total_num_elements].

        Returns:
            prediction associated with each composition [batchsize].
        """
        return self.fc(composition_feas).view(-1)

    def fit(
        self,
        structures_or_graphs: Sequence[Structure | CrystalGraph],
        energies: Sequence[float],
    ) -> None:
        """Fit the model to a list of crystals and energies.

        Args:
            structures_or_graphs (list[Structure  |  CrystalGraph]): Any iterable of
                pymatgen structures and/or graphs.
            energies (list[float]): Target energies.
        """
        num_data = len(energies)
        composition_feas = torch.zeros([num_data, self.max_num_elements])
        e = torch.zeros([num_data])
        for index, (structure, energy) in enumerate(
            zip(structures_or_graphs, energies, strict=True)
        ):
            if isinstance(structure, Structure):
                atomic_number = torch.tensor(
                    [site.specie.Z for site in structure],
                    dtype=torch.int32,
                    requires_grad=False,
                )
            else:
                atomic_number = structure.atomic_number
            composition_fea = torch.bincount(
                atomic_number - 1, minlength=self.max_num_elements
            )
            if self.is_intensive:
                composition_fea = composition_fea / atomic_number.shape[0]
            composition_feas[index, :] = composition_fea
            e[index] = energy

        # Use numpy for pinv
        self.feature_matrix = composition_feas.detach().numpy()
        self.energies = e.detach().numpy()
        state_dict = collections.OrderedDict()
        weight = (
            np.linalg.pinv(self.feature_matrix.T @ self.feature_matrix)
            @ self.feature_matrix.T
            @ self.energies
        )
        state_dict["weight"] = torch.tensor(weight).view(1, 94)
        self.fc.load_state_dict(state_dict)
        self.fitted = True

    def _assemble_graphs(self, graphs: list[CrystalGraph]) -> Tensor:
        """Assemble a list of graphs into one-hot composition encodings
        Args:
            graphs (list[Tensor]): a list of CrystalGraphs
        Returns:
            assembled batch_graph that contains all information for model.
        """
        composition_feas = []
        for graph in graphs:
            composition_fea = torch.bincount(
                graph.atomic_number - 1, minlength=self.max_num_elements
            )
            if self.is_intensive:
                n_atom = graph.atomic_number.shape[0]
                composition_fea = composition_fea / n_atom
            composition_feas.append(composition_fea)
        return torch.stack(composition_feas, dim=0).float()

    def get_site_energies(self, graphs: list[CrystalGraph]) -> list[Tensor]:
        """Predict the site energies given a list of CrystalGraphs.

        Args:
            graphs (List(CrystalGraph)): a list of Crystal Graph to compute

        Returns:
            a list of tensors corresponding to site energies of each graph [batchsize].
        """
        return [
            self.fc.state_dict()["weight"][0, graph.atomic_number - 1]
            for graph in graphs
        ]

    def initialize_from(self, dataset: str) -> None:
        """Initialize pre-fitted weights from a dataset."""
        if dataset in {"MPtrj", "MPtrj_e"}:
            self.initialize_from_MPtrj()
        elif dataset == "MPF":
            self.initialize_from_MPF()
        else:
            raise NotImplementedError(f"{dataset=} not supported yet")

    def initialize_from_MPtrj(self) -> None:  # noqa: N802
        """Initialize pre-fitted weights from MPtrj dataset."""
        state_dict = collections.OrderedDict()
        state_dict["weight"] = torch.tensor(
            [
                -3.4431,
                -0.1279,
                -2.8300,
                -3.4737,
                -7.4946,
                -8.2354,
                -8.1611,
                -8.3861,
                -5.7498,
                -0.0236,
                -1.7406,
                -1.6788,
                -4.2833,
                -6.2002,
                -6.1315,
                -5.8405,
                -3.8795,
                -0.0703,
                -1.5668,
                -3.4451,
                -7.0549,
                -9.1465,
                -9.2594,
                -9.3514,
                -8.9843,
                -8.0228,
                -6.4955,
                -5.6057,
                -3.4002,
                -0.9217,
                -3.2499,
                -4.9164,
                -4.7810,
                -5.0191,
                -3.3316,
                0.5130,
                -1.4043,
                -3.2175,
                -7.4994,
                -9.3816,
                -10.4386,
                -9.9539,
                -7.9555,
                -8.5440,
                -7.3245,
                -5.2771,
                -1.9014,
                -0.4034,
                -2.6002,
                -4.0054,
                -4.1156,
                -3.9928,
                -2.7003,
                2.2170,
                -1.9671,
                -3.7180,
                -6.8133,
                -7.3502,
                -6.0712,
                -6.1699,
                -5.1471,
                -6.1925,
                -11.5829,
                -15.8841,
                -5.9994,
                -6.0798,
                -5.9513,
                -6.0400,
                -5.9773,
                -2.5091,
                -6.0767,
                -10.6666,
                -11.8761,
                -11.8491,
                -10.7397,
                -9.6100,
                -8.4755,
                -6.2070,
                -3.0337,
                0.4726,
                -1.6425,
                -3.1295,
                -3.3328,
                -0.1221,
                -0.3448,
                -0.4364,
                -0.1661,
                -0.3680,
                -4.1869,
                -8.4233,
                -10.0467,
                -12.0953,
                -12.5228,
                -14.2530,
            ]
        ).view([1, 94])
        self.fc.load_state_dict(state_dict)
        self.is_intensive = True
        self.fitted = True

    def initialize_from_MPF(self) -> None:  # noqa: N802
        """Initialize pre-fitted weights from MPF dataset."""
        state_dict = collections.OrderedDict()
        state_dict["weight"] = torch.tensor(
            [
                -3.4654e00,
                -6.2617e-01,
                -3.4622e00,
                -4.7758e00,
                -8.0362e00,
                -8.4038e00,
                -7.7681e00,
                -7.3892e00,
                -4.9472e00,
                -5.4833e00,
                -2.4783e00,
                -2.0202e00,
                -5.1548e00,
                -7.9121e00,
                -6.9135e00,
                -4.6228e00,
                -3.0155e00,
                -2.1285e00,
                -2.3174e00,
                -4.7595e00,
                -8.1742e00,
                -1.1421e01,
                -8.9229e00,
                -8.4901e00,
                -8.1664e00,
                -6.5826e00,
                -5.2614e00,
                -4.4841e00,
                -3.2737e00,
                -1.3498e00,
                -3.6264e00,
                -4.6727e00,
                -4.1316e00,
                -3.6755e00,
                -2.8030e00,
                6.4728e00,
                -2.2469e00,
                -4.2510e00,
                -1.0245e01,
                -1.1666e01,
                -1.1802e01,
                -8.6551e00,
                -9.3641e00,
                -7.5716e00,
                -5.6990e00,
                -4.9716e00,
                -1.8871e00,
                -6.7951e-01,
                -2.7488e00,
                -3.7945e00,
                -3.3883e00,
                -2.5588e00,
                -1.9621e00,
                9.9793e00,
                -2.5566e00,
                -4.8803e00,
                -8.8604e00,
                -9.0537e00,
                -7.9431e00,
                -8.1259e00,
                -6.3212e00,
                -8.3025e00,
                -1.2289e01,
                -1.7310e01,
                -7.5512e00,
                -8.1959e00,
                -8.3493e00,
                -7.2591e00,
                -8.4170e00,
                -3.3873e00,
                -7.6823e00,
                -1.2630e01,
                -1.3626e01,
                -9.5299e00,
                -1.1840e01,
                -9.7990e00,
                -7.5561e00,
                -5.4690e00,
                -2.6508e00,
                4.1746e-01,
                -2.3255e00,
                -3.4830e00,
                -3.1808e00,
                -1.6934e-02,
                -3.6191e-02,
                -1.0842e-02,
                1.3170e-02,
                -6.5371e-02,
                -5.4892e00,
                -1.0335e01,
                -1.1130e01,
                -1.4312e01,
                -1.4700e01,
                -1.5473e01,
            ]
        ).view([1, 94])
        self.fc.load_state_dict(state_dict)
        self.is_intensive = False
        self.fitted = True

    def initialize_from_numpy(self, file_name: str | Path) -> None:
        """Initialize pre-fitted weights from numpy file."""
        atom_ref_np = np.load(file_name)
        state_dict = collections.OrderedDict()
        state_dict["weight"] = torch.tensor(atom_ref_np).view([1, 94])
        self.fc.load_state_dict(state_dict)
        self.is_intensive = False
        self.fitted = True
